(* Copyright (C) 2019, Daniel Wyckoff.*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)
(*Version 0.3 bool2 -- certified -- version 8.4 pl4 Coq.*)
   
Require Import NaryFunctions.
Require Import Description.
Require Import ProofIrrelevance.
Require Import FunctionalExtensionality.
Require Import Equality.

Require Import SetUtilities.
Require Import TypeUtilities.
Require Import TypeUtilities2.
Require Import SetUtilities2.
Require Import FunctionProperties.
Require Import FunctionProperties2.
Require Import FunctionProperties3.
Require Import FunctionAlgebrasBasics.
Require Import EqDec.
Require Import ListUtilities.
Require Import ArithUtilities.
Require Import LogicUtilities.



(*This implementation of 
many-sorted partial algebras is based on 
Reichel's "Structural Induction on Partial Algebras"
(or equivalently, 
the second edition, "Initial Computability, Algebraic 
Specifications, and Partial Algebras."*)

(*In anticipation of homomorphisms and subalgebras,  
we will name the operator symbols after a number in an 
initial segment of the natural numbers to enforce a 
universal interface.  We will always use a finite number of operations 
on our algebras, which will be crucial when we induct over term algebras.  
It's a bit cumbersome, for the same reason that arbitrary S-sorted 
systems of variables (v:X->S) are so convenient.  Term algebras 
can tolerate infinite systems of variables, but not infinite operations.  *)

Definition signature S (num_ops:nat) : Type :=
  seg_fun num_ops (all_tuples S * S).

Definition signature_seg S num_ops : Type :=
  signature (S @) num_ops.

 
Definition unary_sig {S num_ops} (sig:signature S num_ops) : Prop :=
  forall op (pfop:op < num_ops),
    exists (pfe:projT1 (fst (sig op pfop)) = 1),
      projT2 (fst (sig op pfop)) = transfer_nprod_r
                                     pfe
                                     (snd (sig op pfop), tt).



Lemma projT1_fst_sig0_impl :
  forall {S num_ops} (sig:signature S num_ops) op (pfop:op < num_ops)
         (pfe:projT1 (fst (sig op pfop)) = 0),
    (projT2 (fst (sig op pfop))) =
    transfer_dep_r
      pfe tt.
  intros.
  apply (transfer_dep_inj pfe).
  apply nprod0_eq.
Qed.


Definition op_to_fun {S} (A : S -> Type)
           (op:all_tuples S * S) : Type :=
  tuple_prod_all A (fst op) -> A (snd op).

Definition op_to_fun' {S num_ops} (A:S->Type)
           (sig:signature S num_ops)
           i (pfi:i < num_ops) : Type :=
  op_to_fun A (sig i pfi).

Definition op_to_fun_seg {S} (A :S ---> Type)
           (op:all_tuples (S @) * (S @)) : Type :=
  op_to_fun A op.

Definition op_to_fun_seg' {S num_ops} (A :S ---> Type)
           (sig:signature (S @) num_ops)
           i (pfi:i < num_ops) : Type :=
  op_to_fun_seg A (sig i pfi).

Definition op_to_fun_ens {S T} (A:S->Ensemble T) (op:all_tuples S * S) :
  Type :=
  sig_set (tuple_prod_ens A (projT2 (fst op))) -> sig_set (A (snd op)).

Definition op_to_fun_ens' {S T num_ops}
           (A:S->Ensemble T)
           (sig:signature S num_ops)
           i (pfi:i < num_ops) : Type :=
  op_to_fun_ens A (sig i pfi).


Definition op_to_fun_dep {S} {U:S->Type} (A:forall s:S, Ensemble (U s))
           (op:all_tuples S * S) : Type :=
  sig_set (tuple_prod_ens_dep A (projT2 (fst op))) ->
  sig_set (A (snd op)).

Definition op_to_fun_dep' {S num_ops} {U:S->Type}
           (sig:signature S num_ops)
           (A:forall s:S, Ensemble (U s))
           i (pfi:i < num_ops) : Type :=
  op_to_fun_dep A (sig i pfi).

Definition op_to_fun_ens_to_dep {S T}
           {A:S->Ensemble T}
           {op:all_tuples S * S}
           (f:op_to_fun_ens A op) :
  op_to_fun_dep A op (U := fun _:S => T) :=
  fun (t:sig_set (tuple_prod_ens_dep A (projT2 (fst op)))) =>
    let pfi := in_tuple_prod_ens_dep_const_compat A
                                                  (proj1_sig t)
                                                  (proj2_sig t) in
    f (exist
         (fun c =>
            Inn (tuple_prod_ens A (projT2 (fst op))) c) _
         pfi).

Definition op_to_fun_ens_to_dep' {S T num_ops}
           (sig:signature S num_ops)
           (A:S->Ensemble T)
           i (pfi:i < num_ops)
           (f:op_to_fun_ens A (sig i pfi)) : 
  op_to_fun_dep' sig A i pfi :=
  op_to_fun_ens_to_dep f.
 
           
Definition op_to_fun_ens_seg {S T} (A:S ---> Ensemble T)
           (op:all_tuples (S @) * S @) :
  Type :=
  op_to_fun_ens A op.

Definition op_to_fun_ens_seg' {S T num_ops}
           (sig:signature (S @) num_ops)
           (A:S ---> Ensemble T)
           i (pfi:i < num_ops) : 
  Type :=
  op_to_fun_ens_seg A (sig i pfi).


Definition op_to_fun_seg_prod_type
           {n} (f:seg_fun n Type) (U:Type) : Type :=
  seg_fun_prod_type f -> U.


Definition op_to_pfun {S} (A : S -> Type)
           (op:all_tuples S * S)
           (dom:Ensemble (tuple_prod_all A (fst op))) : Type :=
  fun_in_ens dom (A (snd op)).

Definition op_to_pfun' {S num_ops}
           (sig:signature S num_ops) 
           (A : S -> Type)
           i (pfi:i < num_ops)
           (dom:Ensemble (tuple_prod_all A (fst (sig i pfi)))) : Type :=
  op_to_pfun A (sig i pfi) dom.


Definition op_to_pfun_seg {S} (A : S ---> Type)
           (op:all_tuples (S @) * S @)
           (dom:Ensemble (tuple_prod_all A (fst op))) : Type :=
  op_to_pfun A op dom.

Definition op_to_pfun_seg' {S num_ops}
           (sig:signature (S @) num_ops)
           (A : S ---> Type)
           i (pfi:i < num_ops)
           (dom:Ensemble (tuple_prod_all A (fst (sig i pfi)))) : Type :=
  op_to_pfun_seg A (sig i pfi) dom.
 

Definition op_to_pfun_ens {S T}
           (A:S->Ensemble T) (op:all_tuples S * S)
           (dom:Ensemble (T ^ projT1 (fst op)))
           (pfd: Included dom (tuple_prod_ens A (projT2 (fst op)))) : Type :=
  fun_in_ens dom (sig_set (A (snd op))).

  
Definition op_to_pfun_ens' {S T num_ops}
           (A:S->Ensemble T)
           (sig:signature S num_ops)
           op (pfop:op < num_ops)
           (dom:Ensemble
                  (T ^ (projT1 (fst (sig op pfop)))))
           (pfd: Included dom (tuple_prod_ens A (projT2 (fst (sig op pfop)))))
: Type :=
  op_to_pfun_ens A (sig op pfop) dom pfd.



Definition assment {S X} (A:S->Type) (v:X->S) : Type :=
  forall x:X, A (v x).


Definition assment_seg {X S} (A:S ---> Type) (v:X ----> S) : Type :=
  assment A v.

Definition assment_ens {S U X} (A:S->Ensemble U) (v:X->S) :
  X -> Ensemble U :=
  fun x => A (v x).

Definition assment_ens_sig {S U X} (A:S->Ensemble U) (v:X->S) : Type :=
  forall x:X, sig_set (A (v x)).

Definition map_to_assment_ens_sig
           {S U X n} {A:S->Ensemble U} {v:X->S}
           (x:X^n)
           (a:assment_ens_sig A v) : U^n :=
  map_nprod (fun j (pfj:j < n) =>
               proj1_sig (a (nth_prod j pfj x))).


Definition assment_ens_sig_w'
           {S U X n} {A:S->Ensemble U} {v:X->S}
           (a:assment_ens_sig A v)
           (w:X^n) : U^n :=
  map_nprod (fun i (pfi:i < n) =>
               (proj1_sig (a (nth_prod i pfi w)))).


Lemma in_assment_ens_sig_w' :
  forall {S U X n} {A:S->Ensemble U} {v:X->S}
         (a:assment_ens_sig A v)
         (w:X^n),
    Inn (tuple_prod_ens_sys A v w)
        (assment_ens_sig_w' a w).
  intros S U X n A v a w.
  constructor.
  intros i hi.
  unfold tuple_prod_sys, assment_ens_sig_w'.
  rewrite nth_map_nprod.
  apply proj2_sig.
Qed.

Lemma in_assment_ens_sig_w_nth' :
  forall {S U X n} {A:S->Ensemble U} {v:X->S}
         (a:assment_ens_sig A v)
         (w:X^n)
         i
         (pfi:i < n),
    Inn (A (v (nth_prod i pfi w)))
        (nth_prod i pfi (assment_ens_sig_w' a w)).
  intros S U X n A v a w i hi.
  pose proof (in_assment_ens_sig_w' a w) as h1.
  inversion h1 as [h2].
  specialize (h2 i hi).
  unfold tuple_prod_sys in h2; auto.
Qed.


Inductive in_assment_ens_sig
          {S U X} 
          {A:S->Ensemble U} {v:X->S}
          (u:U)
          (a:assment_ens_sig A v) : Prop :=
  assment_ens_sig_in_intro :
    forall x:X,
      u = proj1_sig (a x) ->
      in_assment_ens_sig u a.


Lemma in_assment_ens_sig_compat :
  forall {S U X} 
         {A:S->Ensemble U} {v:X->S}
         (a:assment_ens_sig A v)
         x,
    Inn (A (v x)) (proj1_sig (a x)).
  intros; apply proj2_sig.
Qed.

Definition w_in_assment_ens_sig 
           {S U X n} 
           {A:S->Ensemble U} {v:X->S}
           (u:U^n)
           (a:assment_ens_sig A v) : Prop :=
  forall i (pfi:i < n),
    in_assment_ens_sig (nth_prod i pfi u) a.

  
Lemma w_in_assment_ens_sig_compat : 
  forall {S U X n} 
         {A:S->Ensemble U} {v:X->S}
         (u:U^n)
         (a:assment_ens_sig A v)
         (x:X^n),
    Inn (tuple_prod_ens_sys A v x)
        (map_nprod (fun j (pfj:j < n) =>
                      proj1_sig (a (nth_prod j pfj x)))).
  intros S U X n A v u a x.
  constructor.
  intros j hj.
  unfold tuple_prod_sys.
  rewrite nth_map_nprod.
  apply proj2_sig.
Qed.

Definition in_assment_ens_sig_at_op 
           {S U X num_ops}
           {sig:signature S num_ops}
           {A:S->Ensemble U} {v:X->S}
           op (pfop:op < num_ops)
           (u:U^projT1 (fst (sig op pfop)))
           (a:assment_ens_sig A v) : Prop :=
  w_in_assment_ens_sig u a.

 

Definition assment_dep {X S} {U:S->Type}
           (A:forall s:S, Ensemble (U s))
           (v:X->S) : Type :=
  forall x:X, sig_set (A (v x)).


Definition assment_ens_sig_to
           {S U X} {A:S->Ensemble U} {v:X->S}
           (a:assment_ens_sig A v) :
  @assment_dep X S (fun _=> U) (fun s => (A s)) v :=
  a.


Definition assment_dep_const_to
           {S U X} {A:S->Ensemble U} {v:X->S}
           (a:@assment_dep X S (fun _=> U) (fun s => (A s)) v) :
  assment_ens_sig A v :=
  a.


Definition assment_w {S n} (A:S -> Type)
           (w:S^n) : Type :=
  tuple_prod A w.


Definition assment_w_ens {S U n} (A:S -> Ensemble U)
           (w:S^n) : Type :=
  sig_set (tuple_prod_ens A w).

Definition assment_w_dep
           {S n} {U:S -> Type}
           (A:forall s:S, Ensemble (U s))
           (w:S^n) : Type :=
  sig_set (tuple_prod_ens_dep A w).


Fixpoint in_pre_w {S X n} :
  forall (w:S^n) (v:X->S), Prop :=
  match n with
    | 0 => fun _ _ => True
    | S n' =>
      fun w v =>
        let (fw, sw) := w in
        in_pre fw v /\
        (in_pre_w sw v)
  end.



Definition pre_sys_aux {S X} (v:X->S) (s:S) : Ensemble X :=
  [x:X | v x = s].


 
Lemma assment_w_tt :
  forall {S} (A:S -> Type), assment_w A tt (n:=0) = unit.
  auto.
Qed.

Lemma assment_w_tt_ :
  forall {S} (A:S -> Type) (w:S^0),
    assment_w A tt (n:=0) = unit.
  auto.
Qed.

Lemma assment_w_tt' :
  forall {S U} (A:S -> Ensemble U)
         (w:S^0),
    assment_w_ens A w = sig_set (Singleton tt).
  unfold assment_w_ens; intros; rewrite tuple_prod_ens0; auto.
Qed.

Lemma assment_w_dep_tt :
  forall {S} {U:S -> Type}
         (A:forall s:S, Ensemble (U s))
         (w:S^0),
    assment_w_dep A w =
    sig_set (Singleton tt).
  unfold assment_w_dep; intros; rewrite tuple_prod_ens_dep0; auto.
Qed.


Lemma assment_w_S' :
  forall {T U n} (A:T -> Ensemble U)
         (w:T^(S n)),
    assment_w_ens A w =
    sig_set [u:U^(S n) |
               Inn (A (fst w)) (fst u) /\
               Inn (tuple_prod_ens A (snd w))
                   (snd u)].
  unfold assment_w_ens; intros; rewrite tuple_prod_ens_S;
  destruct w; auto.
Qed.

Lemma assment_w_dep_S :
  forall {T n}
         {U:T -> Type}
         (A:forall t:T, Ensemble (U t))
         (w:T^(S n)),
    assment_w_dep A w =
    sig_set [t:tuple_prod U w |
             Inn (A (fst w)) (fst t) /\
             Inn (tuple_prod_ens_dep A (snd w))
                 (snd t)].
  unfold assment_w_dep; intros; rewrite tuple_prod_ens_dep_S;
  destruct w; auto.
Qed.


Lemma in_assment_w_dep_S :
  forall {T n} {U:T -> Type}
         {A:forall t:T, Ensemble (U t)}
         {w:T^(S n)}
         (p:assment_w_dep A w),
    Inn [t:tuple_prod U w |
             Inn (A (fst w)) (fst t) /\
             Inn (tuple_prod_ens_dep A (snd w))
                 (snd t)]
        (proj1_sig p).
  intros ? ? ? ? w p ; constructor.
  pose proof (tuple_prod_ens_dep_S A w) as h1.
  pose proof (proj2_sig p) as h2; simpl in h2.
  rewrite h1 in h2 at 1.
  destruct h2; auto.
Qed.



Lemma in_assment_w_dep_S' :
  forall {T n} {U:T -> Type}
         {A:forall t:T, Ensemble (U t)}
         {w:T^(S n)}
         (p:sig_set [t:tuple_prod U w |
                     Inn (A (fst w)) (fst t) /\
                     Inn (tuple_prod_ens_dep A (snd w)) (snd t)]),
    Inn (tuple_prod_ens_dep A w)  (proj1_sig p).
  intros ? ? ? ? w p.
  destruct p as [p hp].
  destruct hp as [hp]; simpl.
  destruct hp as [h1 h2].
  constructor.
  intros i hi; destruct i as [|i]; simpl; auto.
  destruct h2; auto.
Qed.

Lemma in_assment_w_dep_snd :
  forall {T n}
         {U:T -> Type}
         {A:forall t:T, Ensemble (U t)}
         {w:T^(S n)}
         (a:assment_w_dep A w),
    Inn (tuple_prod_ens_dep A (snd w))
        (snd (proj1_sig a)).
  intros ? ? ? ? ? a.
  pose proof (in_assment_w_dep_S a) as h1.
  destruct h1 as [h1]; destruct h1; auto.
Qed.

Definition assment_w_dep_to_S
           {T n}
           {U:T -> Type}
           {A:forall t:T, Ensemble (U t)}
           {w:T^(S n)}
           (a:assment_w_dep A w) :
  assment_w_dep A (snd w) :=
  exist (fun c => Inn (tuple_prod_ens_dep A (snd w)) c)
        (snd (proj1_sig a))
        (in_assment_w_dep_snd a).


Lemma transfer_r_assment_w_dep_S :
  forall {T n}
         {U:T->Type}
         {A:forall t, Ensemble (U t)}
         {w:T^(S n)}
         (p:sig_set
              [t:tuple_prod U w |
               Inn (A (fst w)) (fst t) /\
               Inn (tuple_prod_ens_dep A (snd w)) (snd t)]),
    transfer_r (assment_w_dep_S A w) p =
    exist (fun x : tuple_prod U w =>
              (Inn (tuple_prod_ens_dep A w) x))
          (proj1_sig p)
          (in_assment_w_dep_S' _).
  intros T n U A w p.
  rewrite transfer_r_eq_iff.
  apply f_equal_sig_set.
  rewrite <- (tuple_prod_ens_dep_S); auto; simpl.
  unfold proj1_sig; auto.
Qed.

Lemma transfer_assment_w_dep_S :
  forall {T n}
         {U:T->Type}
         {A:forall t, Ensemble (U t)}
         {w:T^(S n)}
         (p:sig_set (tuple_prod_ens_dep A w)),
    transfer (assment_w_dep_S A w) p =
    exist (fun x : tuple_prod U w =>
             Inn [t:tuple_prod U w |
                  Inn (A (fst w)) (fst t) /\
                  Inn (tuple_prod_ens_dep A (snd w)) (snd t)] x)
          (proj1_sig p)
          (in_assment_w_dep_S _).
  intros T n U A w p.
  rewrite transfer_eq_iff.
  apply f_equal_sig_set.
  rewrite <- (tuple_prod_ens_dep_S); auto; simpl.
  unfold proj1_sig; auto.
Qed.



Lemma fst_transfer_assment_w_dep_S:
  forall {T n}
         {U:T->Type}
         {A:forall t, Ensemble (U t)}
         (w:T^(S n))
         (p:sig_set (tuple_prod_ens_dep A w)),
    fst (proj1_sig (transfer (assment_w_dep_S A w) p)) =
    fst (proj1_sig p).
  intros; rewrite transfer_assment_w_dep_S; simpl; auto.
Qed.

Lemma fst_transfer_r_assment_w_dep_S :
  forall {T n}
         {U:T->Type}
         {A:forall t, Ensemble (U t)}
         (w:T^(S n))
         (p:sig_set [t:tuple_prod U w |
                     Inn (A (fst w)) (fst t) /\
                     Inn (tuple_prod_ens_dep A (snd w))
                         (snd t)]),
    fst (proj1_sig (transfer_r (assment_w_dep_S A w) p)) =
    fst (proj1_sig p).
  intros; rewrite transfer_r_assment_w_dep_S; simpl; auto.
Qed.



Definition seg_fun_prod_ens_to_assment
           {S n}
           {U:S->Type}
           (A:forall s, Ensemble (U s))
           (x:S^n)
           (f:seg_fun_depT_nth U x)
           (pfi:forall i (pfi:i < n),
                    Inn (A (nth_prod i pfi x))
                        (f i pfi)) :
  assment_w_dep A x :=
  exist (fun c => Inn (tuple_prod_ens_dep A x) c)
        (seg_fun_depT_nth_to_tuple_prod x f)
        (in_seg_fun_depT_nth_to_tuple_prod _ _ _ pfi).

  
 
Definition assment_w_seg {S n} (A:S ---> Type) (w:(S @)^n) :
  Type :=
  assment_w A w.

Lemma assment_w_seg_tt :
  forall {S} (A:S ---> Type), assment_w_seg A tt (n:=0) = unit.
  auto.
Qed.


Lemma assment_w_seg_tt_ :
  forall {S} (A:S ---> Type) (w:(S @)^0),
    assment_w_seg A w (n:=0) = unit.
  auto.
Qed.


Fixpoint map_assment_w {S n} :
  forall {A B:S -> Type} {w:S^n}
         (f:Sfun A B)
         (p:assment_w A w),
    assment_w B w :=
  match n with
    | 0 => fun A B w f p  => transfer_r (assment_w_tt_ B w) tt
    | S n' => fun A B w f p =>
                let (ft, st) := w in
                let (fp, sp) := p in
                let f' := map_assment_w f in
                (f  _ fp, f' sp)
  end.

Fixpoint map_assment_w_dep
         {S n} {U V:S -> Type} :
  forall {A:forall s:S, Ensemble (U s)}
         {B:forall s:S, Ensemble (V s)}
         {w:S^n}
         (f:Sfun_dep A B)
         (p:assment_w_dep A w),
    assment_w_dep B w :=
  match n with
    | 0 => fun A B w f p =>
             transfer_r (assment_w_dep_tt B w)
                        (exist (fun c => Inn (Singleton tt) c)
                               tt (In_singleton _ tt))
    | S n' => fun A B w f p =>
                let pfe := assment_w_dep_S A w in
                let pfe' := assment_w_dep_S B w in
                let p' := transfer pfe p in
                let p'' := proj1_sig p' in
                let hp'' := proj2_sig p' in
                let pf1 := sat_in1 _ _ _ hp'' in
                let pf2 := sat_in2 _ _ _ hp'' in
                let ff := f _ _ pf1 in
                let a := map_assment_w_dep
                           f
                           (exist (fun c => Inn (tuple_prod_ens_dep A (snd w)) c) _
                                  pf2) (w := snd w) in
                let pff := intro_characteristic_sat
                             (fun t : tuple_prod V w =>
                             Inn (B (fst w)) (fst t) /\
                             Inn (tuple_prod_ens_dep B (snd w)) (snd t)) 
                             (proj1_sig ff, proj1_sig a)
                             (conj (proj2_sig ff)
                                   (proj2_sig a)) in
                let x := exist _ _ pff in
                let x' := transfer_r pfe' x in
                x'
  end.


Lemma proj1_sig_map_assment_w_dep_eq_iff :
  forall {T n}
         {U V:T->Type}
         {A:forall t:T, Ensemble (U t)}
         {B:forall t:T, Ensemble (V t)}
         {w:T^n}
         (a:assment_w_dep A w)
         (f:forall (t:T) (u:U t), Inn (A t) u ->
                                  {v : V t | Inn (B t) v})
         (t:tuple_prod V w),
    proj1_sig (map_assment_w_dep f a) = t <->
    (forall j (pfj: j < n),
       let pfaj := iff1 (in_tuple_prod_ens_dep_nth_iff A (proj1_sig a))
                        (proj2_sig a) j pfj in
       proj1_sig (f _ _ pfaj) =
       nth_tuple_prod t j pfj).
  intros T n U V A B w a f t; simpl; split.
  intros ? j hj; subst.
  destruct a as [t ht].
  destruct ht as [ht].
  generalize dependent U;
    generalize dependent V;
    generalize dependent j;
    revert w.
  induction n as [|n ih]; simpl.
  intros; omega.
  simpl in ih.
  intro w.
  destruct w as [fw sw].
  intro j; destruct j as [|j]; simpl.
  intros hj V B U A t.
  destruct t as [ft st].
  intros ht f; simpl.
  symmetry.
  gterml.
  redterm0 x.
  redterm0 c.
  redterm0 c0.
  fold c1.
  rewrite (fst_transfer_r_assment_w_dep_S (fw, sw) c1) at 1.
  unfold c, c1; simpl.
  repeat f_equal. 
  apply f_equal_dep.
  gterml.
  redterm0 x0.
  redterm0 c2.
  redterm0 c3.
  fold c4.
  rewrite (fst_transfer_assment_w_dep_S (fw, sw) c4) at 1.
  unfold c4; simpl; auto.
  intros hj V B U A t ht f.
  specialize (ih sw j (lt_S_n _ _ hj) V B U A (snd t)).
  hfirst ih.
  intros i hi.
  specialize (ht (S i) (lt_n_S _ _ hi)).
  simpl in ht.
  pose proof (proof_irrelevance _ hi (lt_S_n _ _ (lt_n_S _ _ hi))) as h1.
  rewrite <- h1 in ht; auto.
  specialize (ih hf f).
  terml ih.
  redterm0 x.
  redterm0 c. 
  fold c0 in ih.
  gterml.
  redterm0 x0. 
  redterm0 c1.
  fold c2.
  pose proof (proof_irrelevance _ c0 c2) as h2.
  rewrite <- h2; auto.
  unfold c0.
  pose proof (transfer_r_sig_set_eq
                _ _
                (tuple_prod_ens_dep_S B (fw, sw))
                (assment_w_dep_S B (fw, sw))) as h3.
  gtermr.
  redterm0 y.
  redterm2 y.
  redterm0 c4.
  redterm0 c5.
  redterm0 c6.
  specialize (h3 c7).
  pose proof (proj1_sig_transfer_r_sig_set c7
                (tuple_prod_ens_dep_S B (fw, sw))
                (assment_w_dep_S B (fw, sw))) as h4.
  pose proof (f_equal (@snd _ _ ) h4) as h5.
  terml h5.
  fold x1 c7.
  pose proof (nth_tuple_prod_functional1 x1 (snd (proj1_sig c7)) 
                                         j j
                                         (lt_S_n _ _ hj)
                                         h5
                                         (eq_refl _)) as h6.
  unfold x1 in h6.
  rewrite h6 at 1.
  simpl.
  gtermr.
  redterm1 y0.
  pose proof (proof_irrelevance
                _
                (eq_lt_trans _ _ _ eq_refl (lt_S_n _ _ hj))
                (lt_S_n _ _ hj)) as h7.
  rewrite h7, transfer_dep_eq_refl.
  fold c0.
  rewrite ih at 1.
  repeat f_equal; apply proj1_sig_injective; simpl.
  unfold c7 in h4; simpl in h4.
  pose proof (f_equal (@snd _ _) h4) as h8; simpl in h8.
  termr h8.
  gtermr.
  fold y1 in h3.
  redterm0 y2.
  fold c9.
  f_equal.
  unfold c9.
  redterm0 c9.
  redterm0 c10.
  fold c11.
  rewrite (proj1_sig_transfer_sig_set
                c11
                (tuple_prod_ens_dep_S A (fw, sw))
                (assment_w_dep_S A (fw, sw))) at 1.
  unfold c11; simpl; auto.
  intro h1.
  generalize dependent U;
    generalize dependent V;
    revert w.
  induction n as [|n ih]; simpl.
  intros; apply unit_eq.
  intro w; destruct w as [fw sw]; simpl.
  intros V B t.
  destruct t as [ft st].
  intros U A a f h1.
  specialize (ih sw V B st U A (assment_w_dep_to_S a) f).
  hfirst ih.
  intros j hj.
  specialize (h1 (S j) (lt_n_S _ _ hj)); simpl in h1.
  pose proof (proof_irrelevance _ hj (lt_S_n _ _ (lt_n_S _ _ hj))) as h2.
  rewrite h2, <- h1; auto.
  repeat f_equal; apply proof_irrelevance.
  specialize (ih hf); subst. 
  gterml.
  redterm0 x.
  redterm0 x.
  pose proof (proj1_sig_transfer_sig_set
                c0
                (tuple_prod_ens_dep_S B (fw, sw))
                (assment_w_dep_S B (fw, sw))) as h8. 
  unfold c0 in h8.
  rewrite transfer_undoes_transfer_r in h8; simpl in h8.
  rewrite <- h8.
  f_equal.
  specialize (h1 0 (lt_0_Sn _)); simpl in h1.
  rewrite <- h1.
  f_equal.
  apply f_equal_dep.
  repeat f_equal; auto.
  apply (proj1_sig_transfer_sig_set
           a
           (tuple_prod_ens_dep_S A (fw, sw))
           (assment_w_dep_S A (fw, sw))).
  repeat f_equal.
  unfold assment_w_dep_to_S.
  apply proj1_sig_injective; simpl.
  f_equal.
  apply (proj1_sig_transfer_sig_set
           a
           (tuple_prod_ens_dep_S A (fw, sw))
           (assment_w_dep_S A (fw, sw))).
Qed.

 
Lemma nth_tuple_prod_map_assment_w_dep :
  forall {T n}
         {U:T->Type}
         {V:T->Type}
         {A:forall t:T, Ensemble (U t)}
         {B:forall t:T, Ensemble (V t)}
         {w:T^n}
         (a:assment_w_dep A w)
         (f:forall (t:T) (u:U t),
              Inn (A t) u ->
              {v:V t | Inn (B t) v})
         j (pfj:j < n),
    nth_tuple_prod (proj1_sig (map_assment_w_dep f a)) j pfj =
    proj1_sig (f (nth_prod j pfj w)
                 (nth_tuple_prod (proj1_sig a) j pfj)
                 ((iff1 (in_tuple_prod_ens_dep_nth_iff A (proj1_sig a))
                        (proj2_sig a)) j pfj)).
  intros T n U V A B w a f j hj.
  pose proof (proj1_sig_map_assment_w_dep_eq_iff
                a f
                (proj1_sig (map_assment_w_dep f a))) as h1;
    simpl in h1.
  hl h1; auto.
  rewrite h1 in hl.
  specialize (hl j hj); auto.
Qed.


(*Total sigma-algebra*)
(*Note the many sorts [A_sigma_t] are not necessarily finite.  
This is to account for things like the term algebra.
The usual convention is a finite a number of operations, and 
a countable set of type variables, but I'm allowing a primitive,
arbitrary type of type variables, which eventually in the development 
will be tamed.*)
Record TAlgebra {S num_ops} (sig:signature S num_ops) : Type :=
  {
    A_sigma_t : S -> Type;
    ops_sigma_t : forall i (pf:i < num_ops),
                  op_to_fun' A_sigma_t sig i pf
  }.

Record TAlgebra_ens {S num_ops} U (sig:signature S num_ops) : Type :=
  {
    type_eq_decS_t' : type_eq_dec S;
    A_sigma_t' : S -> Ensemble U;
    ops_sigma_t' : forall i (pf:i < num_ops),
                  op_to_fun_ens' A_sigma_t' sig i pf
  }.

(*This is the central variant [_dep], 
which is the most universal and most primitive, 
and if it seems that the friction between type theory and 
set theory is too unnecessary here, 
consider inductively defined sets like [Term_algebra_ens] further down.*)
Record TAlgebra_dep {S num_ops} (sig:signature S num_ops) : Type :=
  {
    U_dep_t : S -> Type;
    type_eq_decS_t : type_eq_dec S;
    A_dep_t : forall (s:S), Ensemble (U_dep_t s);
    ops_dep_t : forall i (pf:i < num_ops),
                  op_to_fun_dep' sig A_dep_t i pf
  }.


Definition TAlgebra_to_dep {S U num_ops} {sig:signature S num_ops}
           (A:TAlgebra_ens U sig) :
  TAlgebra_dep sig :=
  Build_TAlgebra_dep 
    S num_ops sig
    (fun _:S => U)
    (type_eq_decS_t' _ _ A)
    (A_sigma_t' U sig A)
    (fun i (pfi:i < num_ops) =>
       (op_to_fun_ens_to_dep' sig (A_sigma_t' U sig A)
                              i pfi (ops_sigma_t' U sig A i pfi))).



Lemma unary_pointwise :
  forall {S num_ops}
         (sig:signature S num_ops)
         (A:TAlgebra_dep sig)
         op (pfop:op < num_ops)
         (pfe:projT1 (fst (sig op pfop)) = 1),
    projT2 (fst (sig op pfop)) = transfer_nprod_r pfe (snd (sig op pfop), tt) ->
    pointwise_ind_eq (U_dep_t _ A)
                     (U_dep_t _ A)
                     (projT2 (fst (sig op pfop)))
                     (transfer_nprod_r pfe (snd (sig op pfop), tt)).
  intros S num_ops sig A op hop he he'. red; intros i hi.
  pose proof (lt_eq_trans _ _ _ hi he) as h4.
  apply lt1 in h4; subst.
  repeat f_equal; auto.
Qed.


(*This predicate returns true if a given S-set generates a 
specific point in a Total Algebra.*)
Inductive point_gen {S num_ops} {sig:signature S num_ops}
          {A:TAlgebra_dep sig}
          (E:forall s:S, Ensemble (U_dep_t sig A s))
          (pfincl:forall s:S, Included (E s) (A_dep_t sig A s)) : 
  forall (s:S) (u:U_dep_t sig A s), Inn (A_dep_t sig A s) u -> Prop :=
| point_gen_point :
    forall (s:S) (u:U_dep_t sig A s)
           (pfin:Inn (E s) u), 
      point_gen E pfincl s u (pfincl s _ pfin)
| point_gen_nil_op :
    forall op (pfop:op < num_ops),
      let n := projT1 (fst (sig op pfop)) in
      let w := projT2 (fst (sig op pfop)) in
      let s := snd (sig op pfop) in
      forall (pfeq:projT1 (fst (sig op pfop)) = 0),
        let pfintt := in_transfer_tt_tuple_prod_ens_dep
                        (U_dep_t sig A)
                        (A_dep_t sig A)
                        w pfeq in
        let u := ops_dep_t
                   sig A op pfop
                   (exist
                    (fun c =>
                       Inn (tuple_prod_ens_dep (A_dep_t sig A)
                                               w)
                           c)
                    _
                    pfintt) in
        point_gen E pfincl s (proj1_sig u) (proj2_sig u)
| point_gen_app :
    forall op (pfop:op < num_ops),
      let n := projT1 (fst (sig op pfop)) in
      let w := projT2 (fst (sig op pfop)) in
      let s := snd (sig op pfop) in
      forall (p:tuple_prod (U_dep_t sig A) w)
             (pfintpl:Inn (tuple_prod_ens_dep (A_dep_t sig A) w) p),
        (forall i (pfi:i < (projT1 (fst (sig op pfop)))),
           point_gen E pfincl
                         (nth_prod i pfi (projT2 (fst (sig op pfop))))
                         (nth_tuple_prod p i pfi)
                       (in_nth_tuple_prod_ens_dep pfintpl i pfi)) ->
        let u := ops_dep_t
                   sig A op pfop
                   (exist (fun c => Inn (tuple_prod_ens_dep
                                           (A_dep_t sig A) w) c)
                          p pfintpl) in
        point_gen E pfincl s (proj1_sig u) (proj2_sig u).


Inductive alg_gen {S num_ops} {sig:signature S num_ops}
          {A:TAlgebra_dep sig}
          (E:forall s:S, Ensemble (U_dep_t sig A s))
          (pfincl:forall s:S, Included (E s) (A_dep_t sig A s)) : Prop :=
| alg_gen_intro :
    (forall (s:S) (u:U_dep_t sig A s) (pfin:Inn (A_dep_t sig A s) u),
       point_gen E pfincl s u pfin) -> alg_gen E pfincl.


(*This returns true for something like 
[in_nest _ . . .  _ e (op3 (op2 (op1 e))) 3], 
where [Inn (E _) e].
So these are always built starting with the "generators", 
i.e. [nest_level] is an absolute nest level.
It returns a count from inward to outward of the number of 
left (right) outward parenthesis bounding an element in [E].
Of course there can be many posibilities for constructor 
parameters, due to the algebra not necessarily being free, 
and no explicit term-rewriting system being specified.
[E] is presumably a set of generators, although 
we don't require it, all the more so in light of the wide range of 
notions of free partial algebras.
This will primarily be 
used for the Term Algebra, which is the free algebra on the 
class of total algebras of signature [sig], so we'll be fine.  
See below if you're looking for a relative parenthesis count starting 
from an arbitrary element (not just a "generator").*)
Inductive in_nest_gen_n {S num_ops} {sig:signature S num_ops}
          {A:TAlgebra_dep sig} 
          {E:forall s:S, Ensemble (U_dep_t sig A s)}
          (pfincle:forall s:S, Included (E s) (A_dep_t sig A s)) :
  forall (s t:S)
         (u:U_dep_t _ A s)
         (v:U_dep_t _ A t) 
         (pfu:Inn (A_dep_t _ A s) u)
         (pfv:Inn (A_dep_t _ A t) v)
         (nest_level:nat), Prop :=
| in_nest0 : forall (s:S) (e:U_dep_t _ A s) (pfe:Inn (E s) e),
               let pfe' := pfincle s _ pfe in
               in_nest_gen_n pfincle s s e e pfe' pfe' 0
| in_nestS : forall (s t:S) (a:U_dep_t _ A s) (b:U_dep_t _ A t)
                    (pfa:Inn (A_dep_t _ A s) a)
                    (pfb:Inn (A_dep_t _ A t) b)
                    n
                    (pfinnest: in_nest_gen_n pfincle s t a b pfa pfb n)
                    op (pfop:op < num_ops)
                    (p:tuple_prod (U_dep_t _ A)
                                  (projT2 (fst (sig op pfop))))
                    (pfp:Inn (tuple_prod_ens_dep
                                (A_dep_t _ A)
                                (projT2 (fst (sig op pfop))))
                             p),
               in_tuple_prod (type_eq_decS_t _ A)
                                 b p (A := U_dep_t _ A) ->
               in_nest_gen_n pfincle s (snd (sig op pfop)) a
                       (proj1_sig
                          (ops_dep_t
                             _ A op pfop
                             (exist (fun c => Inn (tuple_prod_ens_dep _ _)
                                                  c)
                                    p pfp)))
                       pfa
                       (proj2_sig _)
                       (Datatypes.S n).         
           

Inductive in_nest_gen {S num_ops} {sig:signature S num_ops}
          {A:TAlgebra_dep sig} 
          {E:forall s:S, Ensemble (U_dep_t sig A s)}
          (pfincle:forall s:S, Included (E s) (A_dep_t sig A s)) :
  forall (s t:S)
         (u:U_dep_t _ A s)
         (v:U_dep_t _ A t) 
         (pfu:Inn (A_dep_t _ A s) u)
         (pfv:Inn (A_dep_t _ A t) v), Prop :=
  | in_nest_gen_intro : 
      forall (s t:S)
             (u:U_dep_t _ A s)
             (v:U_dep_t _ A t) 
             (pfu:Inn (A_dep_t _ A s) u)
             (pfv:Inn (A_dep_t _ A t) v)
             (nest_level:nat),
        in_nest_gen_n pfincle s t u v pfu pfv (Datatypes.S nest_level) ->
        in_nest_gen pfincle s t u v pfu pfv.

(*If this were used in anything but a free algebra (like the term
algebra), it would be almost useless in current form.
This is the relative form of [in_nest_n].*)
Definition in_nest_n_rel {S num_ops} {sig:signature S num_ops}
          {A:TAlgebra_dep sig} 
          (s t:S)
          (u:U_dep_t _ A s)
          (v:U_dep_t _ A t) 
          (pfu:Inn (A_dep_t _ A s) u)
          (pfv:Inn (A_dep_t _ A t) v)
          (nest_level:nat) : Prop :=
  in_nest_gen_n
    (fun c:S => inclusion_reflexive (A_dep_t _ A c)) s t u v pfu pfv nest_level.

Definition in_nest_rel {S num_ops} {sig:signature S num_ops}
          {A:TAlgebra_dep sig} 
          (s t:S)
          (u:U_dep_t _ A s)
          (v:U_dep_t _ A t) 
          (pfu:Inn (A_dep_t _ A s) u)
          (pfv:Inn (A_dep_t _ A t) v)
          (nest_level:nat) : Prop :=
  in_nest_gen (fun c:S => inclusion_reflexive (A_dep_t _ A c)) s t u v pfu pfv.


Lemma in_nest_gen_S_iff :
  forall {S num_ops} {sig:signature S num_ops}
         {A:TAlgebra_dep sig} 
         {E:forall s:S, Ensemble (U_dep_t sig A s)}
         {s t:S}
         (pfds:type_eq_dec S)
         (pfincle:forall s:S, Included (E s) (A_dep_t sig A s))
         (u:U_dep_t _ A s)
         (v:U_dep_t _ A t)
         (pfu:Inn (A_dep_t _ A s) u)
         (pfv:Inn (A_dep_t _ A t) v)
         n,
    in_nest_gen_n pfincle s t u v pfu pfv (Datatypes.S n) <-> 
    (exists op (pfop:op < num_ops)
            (p:tuple_prod (U_dep_t _ A) (projT2 (fst (sig op pfop))))
            (pfp:Inn (tuple_prod_ens_dep _ _) p)
            (t':S) (w:U_dep_t sig A t')
            (pfw:Inn (A_dep_t sig A t') w)
            (pfe:t = snd (sig op pfop)),
       (in_tuple_prod pfds w p /\
        v = transfer_dep_r
              pfe
              (proj1_sig (ops_dep_t _ A op pfop (exist _ p pfp)))) /\
       ((n = 0 /\ Inn (E s) u /\
         exists (pfe':s = t'),
           u = transfer_dep_r pfe' w) \/
        (n > 0 /\
         in_nest_gen_n pfincle s t' u w pfu pfw n))).
  intros S num_ops sig A E s t hds he u v hu hv n; split; intro h1.  
  inversion h1; subst; clear h1.
  clear H6 H7 H5.
  apply existT_injective2 in H4; subst.
  apply existT_injective2 in H; subst.
  exists op, pfop, p, pfp, t0, b, pfb, (eq_refl _).

 
  rewrite transfer_dep_r_eq_refl. 
  split; auto.
  split; auto.
  erewrite type_eq_dec_eq at 1.
  apply H3.
  destruct (zerop n) as [|hn]; subst.
  inversion pfinnest; subst.
  apply existT_injective2 in H9; subst.
  apply existT_injective2 in H7; subst.
  rewrite in_tuple_prod_iff in H3.
  destruct H3 as [i [hi [h9 h10]]]; subst.
  rewrite transfer_dep_eq_refl in h10; subst.  
  left; auto.
  inversion pfinnest; subst.
  apply existT_injective2 in H4; subst; split; auto.
  apply existT_injective2 in H7; subst.
  split; auto.
  exists (eq_refl _). 
  rewrite transfer_dep_r_eq_refl; auto.
  right.
  pose proof (proof_irrelevance _ hu pfa); subst; auto with arith. 
  destruct h1 as [op [hop [p [hp [t' [w' [hw' [he' [h1 h2]]]]]]]]];
    destruct h1 as [h1 h3]; subst.
  revert hv.
  rewrite transfer_dep_r_eq_refl. 
  intro hv.

  destruct h2 as [h2|h2]. 
  destruct h2 as [? [h2 [h3 h4]]]; subst.
  rewrite transfer_dep_r_eq_refl in h2.
  revert hu.
  rewrite transfer_dep_r_eq_refl.
  intro hu.  
  pose proof (in_nest0 he t' w' h2) as h6.
  pose proof (in_nestS he _ _ _ _ _ _ _ h6 op hop p hp) as h7.
  hfirst h7.
  erewrite type_eq_dec_eq at 1.
  apply h1.
  specialize (h7 hf).
  erewrite f_equal_dep.
  apply h7.
  apply proof_irrelevance.
  destruct h2 as [h2 h3].
  pose proof (in_nestS he _ _ _ _ _ _ _ h3 op hop p hp) as h7. 
  hfirst h7.
  erewrite type_eq_dec_eq at 1.
  apply h1.
  specialize (h7 hf).
  erewrite f_equal_dep.
  apply h7.
  auto.
Qed.




    

Definition sub_elts
           {S num_ops} {sig:signature S num_ops}
           {A:TAlgebra_dep sig} 
           {E:forall s:S, Ensemble (U_dep_t sig A s)}
           {pfincle:forall s:S, Included (E s) (A_dep_t sig A s)} 
           (t:S)
           (v:U_dep_t _ A t)
           (pfv:Inn (A_dep_t _ A t) v) : 
  forall (s:S), Ensemble (U_dep_t _ A s) :=
  fun s =>
    [u:U_dep_t _ A s |
     exists (pfu:Inn (A_dep_t _ A s) u),
     in_nest_gen pfincle s t u v pfu pfv].

Definition sub_elts_n 
           {S num_ops} {sig:signature S num_ops}
           {A:TAlgebra_dep sig} 
           {E:forall s:S, Ensemble (U_dep_t sig A s)}
           {pfincle:forall s:S, Included (E s) (A_dep_t sig A s)} 
           (t:S)
           (v:U_dep_t _ A t)
           (pfv:Inn (A_dep_t _ A t) v) n : 
  forall (s:S), Ensemble (U_dep_t _ A s) :=
  fun s => 
    [u:U_dep_t _ A s |
     exists (pfu:Inn (A_dep_t _ A s) u),
     in_nest_gen_n pfincle s t u v pfu pfv n].


(*Total homomorphism between two total algebras.*)
Inductive Thomo {S num_ops} {sig:signature S num_ops}
          {A B:TAlgebra sig} 
          (f:Sfun (A_sigma_t sig A) (A_sigma_t sig B)) : Prop :=
  Thomo_intro : (forall i (pfi:i < num_ops),
                   let w := fst (sig i pfi) in
                   let s := snd (sig i pfi) in
                   let opA := ops_sigma_t sig A i pfi in
                   let opB := ops_sigma_t sig B i pfi in
                   forall (a:assment_w (A_sigma_t _ A) (projT2 w)),
                     let f' := map_assment_w f in
                     f _ (opA a) = opB (f' a)) ->
                Thomo f.

(*Dependent-Ensemble type form -- the most general, and metaphysically 
pure.*)
Inductive Thomo_dep {S num_ops} {sig:signature S num_ops}
          {A B:TAlgebra_dep sig}
          (f:Sfun_dep (A_dep_t sig A) (A_dep_t sig B)) : Prop :=
  Thomo_dep_intro : (forall i (pfi:i < num_ops),
                       let w := fst (sig i pfi) in
                       let s := snd (sig i pfi) in
                       let opA := ops_dep_t sig A i pfi in
                       let opB := ops_dep_t sig B i pfi in
                       forall (a:assment_w_dep (A_dep_t _ A)
                                                 (projT2 w)),
                         let opa := proj1_sig (opA a) in
                         let pfopa := proj2_sig (opA a) in
                         let f' := map_assment_w_dep f in
                         f _ opa pfopa =
                         opB (f' a)) ->
                    Thomo_dep f.


Lemma homos_eq_on_generators :
  forall {S num_ops} {sig:signature S num_ops} {A B:TAlgebra_dep sig}
         (E:forall s:S, Ensemble (U_dep_t sig A s))
         (pfincl:forall s:S, Included (E s) (A_dep_t sig A s))
         (f g:Sfun_dep (A_dep_t sig A) (A_dep_t sig B)),
    Thomo_dep f ->
    Thomo_dep g ->
    alg_gen E pfincl ->
    (forall (s:S)
            (u:U_dep_t sig A s)
            (pfin:Inn (E s) u),
       proj1_sig (f s u (pfincl _ _ pfin)) =
       proj1_sig (g s u (pfincl _ _ pfin))) ->
    (forall op (pfop:op < num_ops),
       let n := projT1 (fst (sig op pfop)) in
       let w := projT2 (fst (sig op pfop)) in
       let s := snd (sig op pfop) in
       forall (pfeq:projT1 (fst (sig op pfop)) = 0),
         let pfintt := in_transfer_tt_tuple_prod_ens_dep
                         (U_dep_t sig A)
                         (A_dep_t sig A)
                         w pfeq in
         let u := ops_dep_t
                    sig A op pfop
                    (exist
                    (fun c =>
                       Inn (tuple_prod_ens_dep (A_dep_t sig A)
                                               w)
                           c)
                    _
                    pfintt) in
         f s (proj1_sig u) (proj2_sig u) = 
         g s (proj1_sig u) (proj2_sig u)) ->
    f = g.
  intros S num_ops sig A B E hincl f g h1 h2 h3 h4 hne.
  destruct h1 as [h1], h2 as [h2].
  apply functional_extensionality_dep.
  intro s.
  apply functional_extensionality_dep.
  intro u.
  apply functional_extensionality_dep.
  intro hu.
  destruct h3 as [h3]. 
  specialize (h3 s u hu). 
  induction h3.
  apply proj1_sig_injective; auto.
  apply h4.
  specialize (h1 op pfop).
  specialize (h2 op pfop).
  simpl in h1, h2.
  specialize (hne op pfop pfeq); simpl in hne.
  pose proof h4 as h4'.
  specialize (h4 s (proj1_sig u)). 
  specialize
    (h1
       (exist (fun c => Inn (tuple_prod_ens_dep (A_dep_t sig A) w) c)
              _ pfintt)).
  terml hne; terml h1.
  assert (he:x = x0).
    unfold x, x0.
    repeat f_equal.
  fold x0 in h1, hne.
  fold x in hne; fold x0 in h1.
  termr hne; gtermr.
  assert (he':y = y0). auto.
  fold y0.
  rewrite <- he'.
  fold y in hne.
  rewrite <- hne.
  unfold x, s,u. 
  repeat f_equal.
  specialize (h1 op pfop);
    specialize (h2 op pfop);
    simpl in h1, h2.
  specialize (h1
                (exist (fun c =>
                          Inn (tuple_prod_ens_dep (A_dep_t sig A) w) c)
                       _ pfintpl)).
  specialize (h2
                
                (exist (fun c =>
                          Inn (tuple_prod_ens_dep (A_dep_t sig A) w) c)
                       _ pfintpl)).
  terml h2.
  redterm1 x.
  unfold u.
  unfold s, w.
  unfold s, w in p, pfintpl, h1, h2, H, H0.
  clear c x u s n w.
  rewrite h1.
  rewrite h2 at 1.
  f_equal.
  apply proj1_sig_injective; simpl.
  pose proof
       (proj1_sig_map_assment_w_dep_eq_iff
          (exist
             (fun c : tuple_prod (U_dep_t sig A)
                                 (projT2 (fst (sig op pfop))) =>
                Inn (tuple_prod_ens_dep (A_dep_t sig A)
                                        (projT2 (fst (sig op pfop)))) c)
                    p pfintpl)
          f
          (proj1_sig
             (map_assment_w_dep
                g
                (exist
                   (fun c : tuple_prod
                                        (U_dep_t sig A)
                                        (projT2 (fst (sig op pfop))) =>
                      Inn
                        (tuple_prod_ens_dep
                           (A_dep_t sig A)
                           (projT2 (fst (sig op pfop)))) c)
                   p pfintpl)))) as h3; simpl in h3.
  rewrite h3.
  intros j hj.
  specialize (H0 j hj).
  terml H0.
  redterm0 x.
  gterml.
  redterm0 x0.
  redterm0 c0.
  pose proof (proof_irrelevance _ c1 c) as h6.
  fold c1.
  fold c in H0.
  rewrite h6, H0.
  rewrite nth_tuple_prod_map_assment_w_dep; simpl.
  repeat f_equal.
  apply proof_irrelevance.
Qed.

                                                           
(*Partial sigma-algebra*)
Record PAlgebra {S num_ops} (sigma:signature S num_ops) : Type :=
  {
    A_sigma : S -> Type;
    doms : forall i (pfi:i < num_ops),
             (Ensemble (tuple_prod_all A_sigma (fst (sigma i pfi))));
    ops_sigma : forall i (pfi:i < num_ops),
                op_to_pfun A_sigma (sigma i pfi) (doms i pfi)
  }.

Record PAlgebra_ens {S num_ops} U (sigma:signature S num_ops) : Type :=
  {
    A_sigma_ens : S -> Ensemble U;
    doms_ens : forall op (pfop:op < num_ops),
                 Ensemble (U^projT1 (fst (sigma op pfop)));
    incl_doms_ens : forall op (pfop:op < num_ops),
                      Included (doms_ens op pfop)
                               (tuple_prod_ens
                                  A_sigma_ens
                                  (projT2 (fst (sigma op pfop))));
    ops_sigma_ens : forall op (pfop:op < num_ops),
                      op_to_pfun_ens A_sigma_ens (sigma op pfop)
                                     (doms_ens op pfop)
                                     (incl_doms_ens op pfop)
  }.



Lemma pointwise_ind_eq_seg_fun_prod_type_op :
  forall {S num_ops}
         {sig:signature S num_ops}
         (A:TAlgebra_dep sig)
         op (pfop:op < num_ops)
         (f:seg_fun_prod_type
              (fun (j : nat) (pfj : j < projT1 (fst (sig op pfop))) =>
                 U_dep_t sig A (nth_prod j pfj (projT2 (fst (sig op pfop)))))),
    pointwise_ind_eq (seg_fun_type_to_ind
                        (fun (j : nat) (pfj : j < projT1 (fst (sig op pfop))) =>
                           U_dep_t sig A (nth_prod j pfj (projT2 (fst (sig op pfop))))))
                     (U_dep_t sig A)
                     (full_segT_ind (projT1 (fst (sig op pfop))))
                     (projT2 (fst (sig op pfop))).
  intros S num_ops sig A op hop f; red; intros i hi.
  unfold seg_fun_type_to_ind.
  f_equal.
  unfold full_segT_ind.
  rewrite nth_map_nprod; simpl; auto.
Qed.


Definition term X num_ops : Type :=
  list (X + (num_ops @)).



Definition pre_sys {S X} num_ops (v:X->S) (s:S) :
  Ensemble (term X num_ops) :=
  Im (pre_sys_aux v s) (fun x => inl x::nil).

Lemma in_pre_sys :
  forall {S X} num_ops (v:X->S) (s:S) x,
    Inn (pre_sys num_ops v (v x))
        (inl x::nil).
  intros; apply Im_intro with x; constructor; auto.
Qed.



(*Prefix Polish Notation.*)
Inductive Term_algebra_ens {S X num_ops}
          (sig:signature S num_ops)
          (v:X->S) :
  S -> Ensemble (term X num_ops) :=
| term_algebra_ensX : forall (x:X),
                    Inn (Term_algebra_ens sig v (v x))
                        ((inl x)::nil)
| term_algebra_nil_op :
    forall i (pfi:i < num_ops),
      projT1 (fst (sig i pfi)) = 0 ->
      Inn (Term_algebra_ens sig v (snd (sig i pfi)))
          (inr (exist (fun c => c < num_ops) i pfi) :: nil)
| term_algebra_ens_app :
    forall i (pfi: i < num_ops)
           (t:(term X num_ops)^projT1 (fst (sig i pfi))),
      (forall j (pfj:j < projT1 (fst (sig i pfi))),
         Inn (Term_algebra_ens
                sig v
                (nth_prod j pfj (projT2 (fst (sig i pfi)))))
             (nth_prod j pfj t)) ->
      Inn (Term_algebra_ens sig v (snd (sig i pfi)))
          (inr (exist (fun c => c < num_ops) i pfi) ::
               app_list (nprod_to_list _ _ t)).



(*I know inductive sets don't need something like this, but oftentimes 
the actual parameters are not in form required by constructors.*)
Lemma in_term_algebra_ens_iff :
  forall {S X num_ops} (sig:signature S num_ops)
         (v:X->S) (s:S)
         (t:term X num_ops),
    Inn (Term_algebra_ens sig v s) t <->
    (exists x,
       s = v x /\
       t = (inl x)::nil) \/
    (exists i (pfi:i < num_ops),
       s = snd (sig i pfi) /\
       projT1 (fst ((sig i pfi))) = 0 /\
       t = inr _ (exist (fun c => c < num_ops) i pfi) :: nil) \/
    (exists i (pfi:i < num_ops)
            (t':(term X num_ops)^projT1 (fst (sig i pfi))),
       s = snd (sig i pfi) /\
       (forall j (pfj:j < projT1 (fst (sig i pfi))),
          Inn (Term_algebra_ens
                 sig v
                (nth_prod j pfj (projT2 (fst (sig i pfi)))))
              (nth_prod j pfj t')) /\
       t = inr (exist (fun c => c < num_ops) i pfi)
               :: (app_list (nprod_to_list _ _ t'))).
  intros S X num_ops sig v s t; split; intro h2. 
  inversion h2; subst.
  left.  
  exists x; auto.
  right; left.
  exists i, pfi.
  split; auto.
  right; right.
  exists i, pfi, t0.
  split; auto. 
  destruct h2 as [h2|[h2|h2]].  
  destruct h2 as [x [h2 ?]]; subst.
  constructor.
  destruct h2 as [i [h2 [h3 [h4 ?]]]]; subst.
  constructor.
  assumption.
  destruct h2 as [i [h2 [t' [? [h3 ?]]]]]; subst.
  constructor; auto.
Qed.

Lemma in_term_algebra_not_nil :
  forall {S X num_ops} (sig:signature S num_ops)
         (v:X->S)
         (t:term X num_ops)
         (s:S),
    Inn (Term_algebra_ens sig v s) t ->
    t <> nil.
  intros ? ? ? ? ? ? ? h1;
  induction h1; apply cons_neq_nil.
Qed.


Lemma in_nth_tuple_prod_term_algebra :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {v:X->S}
         {op} {pfop:op < num_ops} 
         (a':assment_w_dep
               (Term_algebra_ens sig v)
               (projT2 (fst (sig op pfop))))
         j (hj:j < projT1 (fst (sig op pfop)))
         x
         (pfe:nth_prod j hj (projT2 (fst (sig op pfop))) = v x),
    Inn (Term_algebra_ens sig v (v x))
        (nth_tuple_prod (proj1_sig a') j hj).
  intros ? ? ? ? ? ? ? a'.
  destruct a' as [p hp].
  destruct hp as [hp].
  intros j hj; pose proof (hp j hj) as h1; simpl.
  intro x.
  inversion h1; clear h1.
  intro.
  rewrite <- pfe.
  constructor.
  intro.  
  rewrite <- pfe.
  constructor; auto.
  intro h1.
  rewrite <- h1.
  constructor.
  intros k hk.
  apply H1.
Qed.




Definition Term_algebra_at_op
           {S X num_ops} (sig:signature S num_ops)
           (v:X->S) op (pfop:op < num_ops) :
  Ensemble ((term X num_ops)^(projT1 (fst (sig op pfop))))
  :=
  tuple_prod_ens (Term_algebra_ens sig v)
                 (projT2 (fst (sig op pfop))).


Definition Term_algebra_ens_op 
           {S X num_ops} (sig:signature S num_ops)
           (v:X->S) op (pfop:op < num_ops) :
  Ensemble (term X num_ops) :=
  Term_algebra_ens sig v (snd (sig op pfop)).



Definition Term_algebra_ens_dep
           {S X num_ops n} (sig:signature S num_ops)
           (v:X->S)
           (w:S^n) :
  Ensemble (tuple_prod (fun _:S => term X num_ops) w) :=
  tuple_prod_ens_dep (Term_algebra_ens sig v) w.


Definition Term_algebra_ens_dep_at_op
           {S X num_ops} (sig:signature S num_ops)
           (v:X->S)
           op (pfop:op < num_ops) :
  Ensemble
    (tuple_prod (fun _:S => term X num_ops) (projT2 (fst (sig op pfop)))) :=
  tuple_prod_ens_dep (Term_algebra_ens sig v)
                     (projT2 (fst (sig op pfop))).


Definition Term_algebra_w 
           {S X num_ops n} (sig:signature S num_ops)
           (v:X -> S) (w:S^n) : Ensemble ((term X num_ops)^n) :=
  tuple_prod_ens (Term_algebra_ens sig v) w.



Definition Term_algebra_sys
           {S X T num_ops} (sig:signature S num_ops)
           (v:X -> S) (w:T->S) : T -> Ensemble (term X num_ops) :=
  tuple_prod_sys (Term_algebra_ens sig v) w.


Definition Term_algebra_op_aux
           {S X num_ops} (sig:signature S num_ops)
           (v:X -> S)
           op (pfop:op < num_ops) :
  fun_in_ens
    (Term_algebra_at_op sig v op pfop)
    (term X num_ops) :=
  fun w pf =>
    (inr (exist (fun c => c < num_ops) op pfop) ::
         app_list (nprod_to_list _ _ w)).


Lemma in_term_algebra_op_aux : 
  forall {S X num_ops} (sig:signature S num_ops)
         (v:X -> S)
         op (pfop:op < num_ops)
         (w:term X num_ops ^ projT1 (fst (sig op pfop)))
         (pf:Inn (Term_algebra_at_op sig v op pfop) w),
    Inn (Term_algebra_ens sig v (snd (sig op pfop)))
        (Term_algebra_op_aux sig v op pfop w pf). 
  intros S X num_ops sig v i hi w hw. 
  unfold Term_algebra_op_aux.
  constructor.
  intros j hj.
  destruct hw as [h1]; subst; auto.
Qed.

Definition Term_algebra_op
           {S X num_ops} (sig:signature S num_ops)
           (v:X -> S)
           op (pfop:op < num_ops) :
  sig_set (Term_algebra_at_op sig v op pfop) ->
  sig_set (Term_algebra_ens sig v (snd (sig op pfop))) :=
  fun x' =>
    let (x, pfx) := x' in
    let y := Term_algebra_op_aux sig v op pfop x pfx in
    let pfy := in_term_algebra_op_aux sig v op pfop x pfx in
    exist _ y pfy.


Definition Term_algebra
           {S X num_ops} (pfds:type_eq_dec S) (sig:signature S num_ops)
           (v:X -> S) : TAlgebra_ens (term X num_ops) sig :=
  Build_TAlgebra_ens S num_ops (term X num_ops) sig
                     pfds
                     (Term_algebra_ens sig v)
                     (Term_algebra_op sig v).


Definition op_term_to_term 
           {S X num_ops} {sig:signature S num_ops}
           {v:X->S} {op} {pfop:op < num_ops}
           (x:sig_set (Term_algebra_at_op sig v op pfop)) : 
  term X num_ops :=
  app_list (nprod_to_list _ _ (proj1_sig x)). 



Definition op_term_to_app_term 
           {S X num_ops} {sig:signature S num_ops}
           {v:X->S} {op} {pfop:op < num_ops}
           (x:sig_set (Term_algebra_at_op sig v op pfop)) :
  term X num_ops :=
  (inr (exist (fun c => c < num_ops) _ pfop))::op_term_to_term x.

Lemma op_term_to_app_term_neq :
  forall {S X num_ops} {sig:signature S num_ops}
         {v:X->S} {op} {pfop:op < num_ops}
         (y:sig_set (Term_algebra_at_op sig v op pfop))
         (x:X),
    op_term_to_app_term y <> inl x::nil.
  unfold op_term_to_app_term; intros; intro; discriminate.
Qed.

           
Lemma in_op_term_to_app_term :
  forall {S X num_ops} {sig:signature S num_ops}
         {v:X->S} {op} {pfop:op < num_ops}
         (x:sig_set (Term_algebra_at_op sig v op pfop)),
    Inn (Term_algebra_ens sig v (snd (sig op pfop)))
        (op_term_to_app_term x).
  unfold op_term_to_app_term;
  intros S X num_ops sig v hop h1 x; destruct x as [x hx];
  constructor; intros j hj; inversion hx; auto.
Qed.

  
Lemma op_term_to_term_sig_tuple_prod_ens_dep_const_to_term_algebra :
  forall {S X num_ops} {sig:signature S num_ops}
         (v:X->S)
         op (pfop: op < num_ops)
           (t:term X num_ops ^ projT1 (fst (sig op pfop)))
           (pft:forall j (pfj:j < projT1 (fst (sig op pfop))),
                  Inn (Term_algebra_ens
                         sig
                         v (nth_prod j pfj (projT2 (fst (sig op pfop)))))
                      (nth_prod j pfj t)),
    op_term_to_term (sig_tuple_prod_ens_dep_const_to
                       (Term_algebra_ens sig v)
                       (projT2 (fst (sig op pfop)))
                         (seg_fun_prod_ens_to_assment
                            (Term_algebra_ens sig v)
                            (projT2 (fst (sig op pfop)))
                            (fun j (pfj:j < projT1 (fst (sig op pfop))) =>
                               nth_prod j pfj t) pft)) =
    app_list
      (nprod_to_list
         (term X num_ops) (projT1 (fst (sig op pfop))) t).
  unfold op_term_to_term, seg_fun_prod_ens_to_assment; intros; f_equal.
  pose proof
       (sig_tuple_prod_ens_dep_const_to_seg_fun_depT_nth_to_tuple_prod
          (Term_algebra_ens sig v)
          (projT2 (fst (sig op pfop)))
          t pft) as h1.
  f_equal.
  simpl in h1; simpl; auto.
Qed.


Definition Term_algebra_op_aux_dep
           {S X num_ops} (sig:signature S num_ops)
           (v:X -> S)
           i (pfi:i < num_ops) :
  fun_in_ens
    (Term_algebra_ens_dep_at_op sig v i pfi)
    (term X num_ops) :=
  fun p pfp =>
    let pfp' := in_tuple_prod_ens_dep_const_compat _ _ pfp in
    op_term_to_app_term (exist _ _ pfp').


Lemma in_term_algebra_op_aux_dep : 
  forall {S X num_ops}
         (sig:signature S num_ops)
         (v:X -> S)
         i (pfi:i < num_ops)
         (p:tuple_prod (fun _:S => term X num_ops)
                       (projT2 (fst (sig i pfi))))
         (pfp:Inn (Term_algebra_ens_dep_at_op sig v i pfi)
                  p),
    let pfp' := in_tuple_prod_ens_dep_const_compat _ _ pfp in
    Inn (Term_algebra_ens sig v (snd (sig i pfi)))
        (op_term_to_app_term (exist _ _ pfp')).
  unfold op_term_to_app_term;
  intros S X num_ops sig v i hi p hp; simpl.
  constructor; simpl.
  intros; inversion hp; auto.
  specialize (H j pfj).
  rewrite nth_prod_transfer_tuple_prod_const; auto.
Qed.

Definition Term_algebra_dep'
           {S X num_ops} (sig:signature S num_ops)
           (v:X -> S)
           i (pfi:i < num_ops) :
  op_to_fun_dep' sig (Term_algebra_ens sig v) i pfi :=
  fun (x':sig_set (Term_algebra_ens_dep_at_op sig v i pfi)) =>
    let (x, pfx) := x' in
    let y := Term_algebra_op_aux_dep sig v i pfi x pfx in
    let pfx' := in_term_algebra_op_aux_dep sig v i pfi x pfx in
    exist (fun c => Inn (Term_algebra_ens sig v (snd (sig i pfi))) c)
          y pfx'.


Definition Term_algebra_dep {S X num_ops}
           (pfds:type_eq_dec S) 
           (sig:signature S num_ops) (v:X->S)  :
  TAlgebra_dep sig :=
  Build_TAlgebra_dep S num_ops sig (fun s => term X num_ops)
                     pfds
                     (Term_algebra_ens sig v)
                     (Term_algebra_dep' sig v).

Lemma incl_term_algebra :
  forall {S X num_ops}
         (pfds:type_eq_dec S) 
         (sig:signature S num_ops) (v:X->S)
         s,
    (Included (Term_algebra_ens sig v s)
              (A_dep_t _ (Term_algebra_dep pfds sig v) s)).
  intros; red; intros; simpl; auto.
Qed.



  
Lemma in_assment_w_dep_term_algebra_ens :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {v:X->S}
         op
         (pfop:op < num_ops)
         (a:assment_w_dep (Term_algebra_ens sig v)
                          (projT2 (fst (sig op pfop)))),
    Inn (Term_algebra_ens_dep_at_op sig v op pfop)
        (proj1_sig a).
  unfold Term_algebra_ens_dep_at_op; intros; constructor; intros i hi.
  pose proof (proj2_sig a) as h1; simpl in h1.
  inversion h1; auto.
Qed.


Lemma proj1_sig_term_algebra_dep_eq' :
  forall {S X num_ops} {sig:signature S num_ops} {A:TAlgebra_dep sig}
         {v:X->S} op (pfop:op < num_ops)
         (a:assment_w_dep (Term_algebra_ens sig v)
                           (projT2 (fst (sig op pfop)))),
    proj1_sig (Term_algebra_dep' sig v op pfop a) = 
    inr (exist (fun c : nat => c < num_ops) op pfop)
        :: op_term_to_term
        (sig_tuple_prod_ens_dep_const_to
           (Term_algebra_ens sig v) (projT2 (fst (sig op pfop))) a).
  intros S X num_ops sig A v op hop a.
  unfold
    Term_algebra_dep',
  Term_algebra_op_aux_dep,
  op_term_to_app_term.
  destruct a; auto.
Qed.

Lemma incl_pre_sys :
  forall {S X} {num_ops} (pfds:type_eq_dec S) (sig:signature S num_ops)
         (v:X->S) s,
    Included (pre_sys num_ops v s)
             (A_dep_t _ (Term_algebra_dep pfds sig v) s).
  intros S X num_ops hds sig v s; red; intros t ht. 
  destruct ht as [x ht]; subst.
  destruct ht as [ht]; subst.
  constructor.
Qed.


Lemma pre_sys_generates_term_algebra' :
  forall {S X num_ops}
         (pfds:type_eq_dec S) 
         (sig:signature S num_ops) (v:X->S),
    alg_gen (pre_sys num_ops v) (incl_pre_sys pfds sig v)
            (A:=Term_algebra_dep pfds sig v). 
  intros S X num_ops hds sig v. 
  constructor.
  intros s u h1.
  pose proof h1 as h2.
  pose proof (point_gen_point (pre_sys num_ops v) (A := Term_algebra_dep hds sig v)) as h3.
  induction h2.
  pose proof (proof_irrelevance _ h1 (term_algebra_ensX sig v x)); subst.
  hfirst h3.
  intros s u h1.
  destruct h1; destruct H; subst; simpl; constructor.
  specialize (h3 hf (v x) (inl x ::nil)).
  hfirst h3.
  unfold pre_sys.
  apply Im_intro with x.
  constructor; auto.
  reflexivity.
  specialize (h3 hf0).
  pose proof (proof_irrelevance _
                                (hf (v x) (inl x::nil) hf0)
                                (term_algebra_ensX sig v x)) as h4.
  pose proof (proof_irrelevance _ hf (incl_pre_sys hds sig v)) as h5; subst.
  rewrite <- h4; auto.
  simpl in h1.
  hfirst h3. 
  intros s u h4. 
  destruct h4; destruct H; destruct H0; subst; simpl; constructor.
  pose proof (point_gen_nil_op (pre_sys num_ops v) (incl_pre_sys hds sig v)
                               (A := Term_algebra_dep hds sig v) i pfi H) as h4; simpl in h4.
  unfold Term_algebra_op_aux_dep, op_term_to_app_term, op_term_to_term in h4; simpl in h4.
  term1 h4.
  redterm0 c.
  assert (h5:c0 = nil).
    unfold c0.
    rewrite app_list_eq_nil_iff.
    left.
    pose proof @nprod_to_list_eq_nil_iff.
    gterml.
    redterm0 x.
    pose proof (nprod_to_list_eq_nil_iff c1) as h6.
    fold c1.
    rewrite h6 at 1; auto.
  pose proof (f_equal (fun c => inr (exist (fun q => q < num_ops) i pfi) :: c)
                        h5) as h6; simpl in h6.
  term0 h4.
  fold c1; fold c1 in h4.
  pose proof c1 as h7; simpl in h7.
  unfold Term_algebra_op_aux_dep, op_term_to_app_term, op_term_to_term in h7; simpl in h7, c1.
  fold c0 in h7.
  rewrite h6 in h7.
  pose proof (subsetT_eq_compat _ _ _ _ c1 h7 h6) as h8.
  unfold Term_algebra_op_aux_dep, op_term_to_app_term, op_term_to_term in h8; simpl in h8.
  dependent rewrite -> h8 in h4.
  pose proof (proof_irrelevance _ h7 h1); subst; auto.
  pose proof (point_gen_app (pre_sys num_ops v) (A := Term_algebra_dep hds sig v)) as h8.
  hfirst h8.
  intros s q hq.
  destruct hq as [q]; subst.
  destruct H1; subst.
  simpl.
  constructor.
  specialize (h8 hf i pfi); simpl in h8.
  specialize (h8 (transfer_to_tuple_prod_const _ t)).
  hfirst h8.
  constructor.
  intros j hj.
  rewrite nth_tuple_prod_transfer_to_tuple_prod_const; auto.
  specialize (h8 hf0).
  hfirst h8.
  intros j hj.
  gen0.
  rewrite nth_tuple_prod_transfer_to_tuple_prod_const.
  intro h9.
  specialize (H0 j hj h9); auto.
  pose proof (proof_irrelevance _ hf (incl_pre_sys hds sig v)); subst; auto.
  specialize (h8 hf1).
  pose proof (proof_irrelevance _ hf (incl_pre_sys hds sig v)); subst; auto.
  term0 h8; simpl in c.
  unfold Term_algebra_op_aux_dep, op_term_to_app_term, op_term_to_term in h8; simpl in h8.
  pose proof (transfer_tuple_prod_const_to_compat (projT2 (fst (sig i pfi))) t) as h9.
  pose proof
       (f_equal
          (fun c =>
             (inr (exist (fun c0 : nat => c0 < num_ops) i pfi)
                  :: app_list
                  (nprod_to_list
                     (term X num_ops) (projT1 (fst (sig i pfi))) c))) h9) as h10; simpl in h10.
  term0 h8.
  pose proof c0 as h11.
  simpl in h11.
  unfold Term_algebra_op_aux_dep, op_term_to_app_term, op_term_to_term in h11; simpl in h11.
  rewrite h10 in h11.
  pose proof (subsetT_eq_compat _ _ _ _ c0  h11 h10) as h12.
  unfold Term_algebra_op_aux_dep, op_term_to_app_term, op_term_to_term in h12; simpl in h12.
  fold c0 in h8.
  dependent rewrite -> h12 in h8.
  pose proof (proof_irrelevance _ h1 h11).
  unfold c0 in h12.
  clear c c0 h9.
  subst; auto.
Qed.


Lemma in_nest_gen_n_var_iff :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {v:X->S}
         (pfds:type_eq_dec S) 
         x y
         (pfx:Inn (Term_algebra_ens sig v (v x)) (inl x::nil))
         (pfy:Inn (Term_algebra_ens sig v (v y)) (inl y::nil))
         n,
    in_nest_gen_n (incl_pre_sys pfds sig v)
                  (v x)
                  (v y)
                  (inl x::nil)
                  (inl y::nil)
                  pfx pfy n
                  (A := Term_algebra_dep pfds sig v) <->
    n = 0 /\ x = y.
  intros S X num_ops sig v hds x y hx hy n; split; intro h1.
  inversion h1; subst.
  inversion H2; tauto.
  destruct h1; subst.
  pose proof (in_nest0 (incl_pre_sys hds sig v)
                       (v y)
                       (inl y::nil)
                       (in_pre_sys _ _ (v y) _)
                       (A := Term_algebra_dep hds sig v)) as h1;
    simpl in h1.
  term1 h1.
  pose proof (proof_irrelevance _ c hx) as he.
  rewrite <- he; clear he.
  pose proof (proof_irrelevance _ c hy) as he.
  rewrite <- he; clear he.
  unfold c; auto.
Qed.


Lemma in_nest_gen_discriminate :
  forall {S X num_ops} {sig:signature S num_ops}
         {v:X->S}
         (pfds:type_eq_dec S)
         op (pfop:op < num_ops)
         x
         n
         (pfr: Inn (Term_algebra_ens sig v (snd (sig op pfop)))
                   (inr (exist (fun c : nat => c < num_ops) op pfop)
                        :: nil)),
    ~ in_nest_gen_n (incl_pre_sys pfds sig v) (v x) 
      (snd (sig op pfop)) (inl x :: nil)
      (inr (exist (fun c : nat => c < num_ops) op pfop) :: nil)
      (term_algebra_ensX sig v x) pfr n
      (A := Term_algebra_dep pfds sig v).
  intros; intro h1.
  dependent induction h1. 
  rename IHh1 into ih.
  destruct pfb.
  rewrite in_nest_gen_n_var_iff in h1. 
  destruct h1; subst.
  pose proof (proof_irrelevance _ pfop pfop0); subst.
  clear x1.
  rename x into hpfe.
  rename H into h1. 
  rename x2 into heqn.
  rewrite in_tuple_prod_iff in h1.
  destruct h1 as [i [h1 [h2 h3]]].
  unfold op_term_to_term in heqn.
  rewrite app_list_eq_nil_iff in heqn; simpl in heqn.
  destruct heqn as [h4|h4].
  rewrite (nprod_to_list_eq_nil_iff (transfer_tuple_prod_const p)) in h4 at 1.
  omega.  
  red in h4.
  specialize (h4 (inl x3::nil)).
  pose proof (f_equal (transfer_dep_r h2) h3) as h5.
  rewrite transfer_dep_r_undoes_transfer_dep in h5.
  rewrite h5 in h4 at 1.
  hfirst h4.
  unfold term.
  erewrite (in_transfer_dep_r_nprod_to_list_const_iff h2 (nth_tuple_prod p i h1)).
  rewrite <- (nth_prod_transfer_tuple_prod_const p).
  apply in_nth_prod.
  specialize (h4 hf); discriminate.
  pose proof (proof_irrelevance _ pfop pfop0); subst.
  specialize (ih i pfi x0). 
  hfirst ih.
  constructor; auto.
  specialize (ih hf).
  hfirst ih.
  constructor.
  specialize (ih hf0).
  hfirst ih.
  constructor.
  specialize (ih hf1 (eq_refl _) (eq_refl _)).
  hfirst ih.
  constructor.
  specialize (ih hf2).
  hfirst ih.
  constructor.
  specialize (ih hf3).
  hfirst ih.
  constructor.
  specialize (ih hf4).
  hfirst ih.
  apply eq_JMeq.
  apply proof_irrelevance.
  specialize (ih hf5); auto.

  rename x2 into h2.
  rename H into h3.
  unfold op_term_to_term in h2.
  simpl in h2.
  rewrite app_list_eq_nil_iff in h2.
  rewrite in_tuple_prod_iff in h3.
  destruct h3 as [j [hj [h4 h5]]].
  destruct h2 as [h2|h2].
  rewrite (nprod_to_list_eq_nil_iff (transfer_tuple_prod_const p)) in h2.
  pose proof (proof_irrelevance _ pfop pfop0); subst; omega.
  red in h2.
  specialize (h2 (nth_tuple_prod p j hj)).
  hfirst h2.
  rewrite <- in_nprod_iff',
  <- (nth_prod_transfer_tuple_prod_const p).
  apply in_nth_prod.
  specialize (h2 hf).
  pose proof pfp as hp.
  destruct hp as [hp].
  specialize (hp j hj).
  simpl in hp.
  rewrite h2 in hp at 1.
  apply in_term_algebra_not_nil in hp.
  contradict hp; auto.
Qed.


Inductive Recursion_homo_aux_var
          {S X num_ops}
          {sig:signature S num_ops}
          {A:TAlgebra_dep sig} {v:X->S}
          (a:assment_dep (A_dep_t sig A) v) : 
  forall (s:S) (t:term X num_ops)
         (u:U_dep_t sig A s)
         (pfu:Inn (A_dep_t sig A s) u),
    Prop :=
  recursion_homo_aux_var_intro :
    forall x, 
      Recursion_homo_aux_var a (v x)
                             (inl x :: nil)
                             (proj1_sig (a x))
                             (proj2_sig _).

Lemma Recursion_homo_aux_var_functional :
  forall {S X num_ops} {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v)
         (s s':S) (t t':term X num_ops)
         (u:U_dep_t sig A s)
         (u':U_dep_t sig A s')
         (pfu: Inn (A_dep_t sig A s) u)
         (pfu':Inn (A_dep_t sig A s') u')
         (pfe:s = s'),
    t = t' ->
    u = transfer_dep_r pfe u' ->
    Recursion_homo_aux_var a s t u pfu ->
    Recursion_homo_aux_var a s' t' u' pfu'.
  intros; subst.
  pose proof (proof_irrelevance _ pfu pfu');
  subst; auto.
Qed.


(*This "operator variant" exists
to enable better abstraction of formal parameters, 
i.e. fail-safe inversion, 
i.e. one that uses primitive inductive parameters in the 
actual parameters --  
so it seems, and is verbose, especially in the second constructor.
If you would a prefer a simpler (concise) version that fails 
on some inversions, then consult the repository tree for just that!*)

Inductive Recursion_homo_aux_op
          {S X num_ops}
          {sig:signature S num_ops}
          {A:TAlgebra_dep sig} {v:X->S}
          (a:assment_dep (A_dep_t sig A) v) 
          op (pfop:op < num_ops) : 
  forall (t:term X num_ops)
         (u:U_dep_t sig A (snd (sig op pfop)))
         (pfu:Inn (A_dep_t sig A (snd (sig op pfop))) u),
    Prop :=
| recursion_homo_aux_op_nil_op_intro :
    let n := projT1 (fst (sig op pfop)) in
    let w := projT2 (fst (sig op pfop)) in
    let s := snd (sig op pfop) in
    forall (pfeq:projT1 (fst (sig op pfop)) = 0),
      let pfintt := in_transfer_tt_tuple_prod_ens_dep
                        (U_dep_t sig A)
                        (A_dep_t sig A)
                        w pfeq in
      let u := ops_dep_t
                 sig A op pfop 
                 (exist
                    (fun c =>
                       Inn (tuple_prod_ens_dep (A_dep_t sig A)
                                               w)
                           c)
                    _
                      pfintt) in
      Recursion_homo_aux_op
        a op pfop
          (inr (exist (fun c => c < num_ops) op pfop) :: nil)
          (proj1_sig u)
          (proj2_sig u)
| rec_homo_aux_op_app :
    let n := projT1 (fst (sig op pfop)) in
    let w := projT2 (fst (sig op pfop)) in
    let s := snd (sig op pfop) in
    forall (ttm:assment_w_dep (Term_algebra_ens
                                 sig v) w)
           (ta:assment_w_dep (A_dep_t sig A) w),
      (forall j (pfj:j < n),
         let wj := nth_prod j pfj w in
         let jttm := nth_tuple_prod (proj1_sig ttm) j pfj in
         let jta := nth_tuple_prod (proj1_sig ta) j pfj in
         let pftaj :=
             iff1 (in_tuple_prod_ens_dep_nth_iff (A_dep_t sig A) _)
                  (proj2_sig ta) j pfj in
         (Recursion_homo_aux_var
            a wj
            (nth_tuple_prod (proj1_sig ttm) j pfj)
            (nth_tuple_prod (proj1_sig ta) j pfj)
            pftaj \/
          exists (pfjttm:jttm <> nil)
                 op' (pfop':op' < num_ops)
                 (pfe:wj = snd (sig op' pfop')),
              let jta' := transfer_dep pfe jta in
              let pftaj' := in_transfer_dep pfe _ pftaj in
              hd' jttm pfjttm =
              inr (exist (fun c => c < num_ops) op' pfop') /\
              Recursion_homo_aux_op a op' pfop' jttm jta' pftaj')) ->
         let opta := ops_dep_t
                       sig A op pfop
                       (exist
                       (fun c =>
                          Inn
                            (tuple_prod_ens_dep (A_dep_t sig A) w) c)
                       (proj1_sig ta)
                       (proj2_sig ta)) in
         let ttm' := sig_tuple_prod_ens_dep_const_to _ _ ttm in
         let tm := inr (exist (fun c => c < num_ops) op pfop) ::
                       (op_term_to_term ttm') in
      Recursion_homo_aux_op a op pfop tm
                            (proj1_sig opta)
                            (proj2_sig opta).

Lemma recursion_homo_aux_op_functional : 
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v) 
         op op'
         (pfop:op < num_ops)
         (pfop':op' < num_ops) 
         (t t':term X num_ops)
         (u:U_dep_t sig A (snd (sig op pfop)))
         (u':U_dep_t sig A (snd (sig op' pfop')))
         (pfu:Inn (A_dep_t sig A (snd (sig op pfop))) u)
         (pfu':Inn (A_dep_t sig A (snd (sig op' pfop'))) u')
         (pfope:op = op'),
    t = t' -> 
    u = transfer_seg_fun_depT2_r
          pfope pfop u'
          (f := fun i pfi => snd (sig i pfi)) ->
    Recursion_homo_aux_op a op pfop t u pfu ->
    Recursion_homo_aux_op a op' pfop' t' u' pfu'.
  intros; subst.
  pose proof (proof_irrelevance _ pfop pfop'); subst.
  revert H1. 
  Gen0.
  rewrite transfer_seg_fun_depT2_r_eq_refl; auto.
  intros h1 h2.
  pose proof (proof_irrelevance _ h1 pfu'); subst; auto.
Qed.

         
Lemma recursion_homo_aux_op_hd : 
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v) 
         op (pfop:op < num_ops) 
         (t:term X num_ops)
         (u:U_dep_t sig A (snd (sig op pfop)))
         (pfu:Inn (A_dep_t sig A (snd (sig op pfop))) u),
    Recursion_homo_aux_op a op pfop t u pfu ->
    exists pf:t <> nil,
      hd' t pf = inr (exist (fun c => c < num_ops) op pfop).
  intros ? ? ? ? ? ? ? ? ? ? ? ? h1.
  destruct h1; try unfold tm; 
  exists (cons_neq_nil _ _); simpl; auto.
Qed.


Inductive Recursion_homo_aux
          {S X num_ops}
          {sig:signature S num_ops}
          {A:TAlgebra_dep sig} {v:X->S} 
          (a:assment_dep (A_dep_t sig A) v) : 
  forall  (s:S) (t:term X num_ops)
          (u:U_dep_t sig A s)
          (pfu:Inn (A_dep_t sig A s) u),
    Prop :=
  recursion_homo_aux_var :
    forall x,
      Recursion_homo_aux a
                          (v x)
                          (inl x :: nil)
                          (proj1_sig (a x))
                          (proj2_sig _)
| recursion_homo_aux_op :
  forall op (pfop:op < num_ops)
         (t:term X num_ops)
         (u:U_dep_t sig A (snd (sig op pfop)))
         (pfu:Inn (A_dep_t sig A (snd (sig op pfop))) u),
    Recursion_homo_aux_op 
      a op pfop t u pfu ->
    Recursion_homo_aux a (snd (sig op pfop))
                        t u pfu.

 
Lemma Recursion_homo_aux_functional :
  forall {S X num_ops} {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v)
         (s s':S) (t t':term X num_ops)
         (u:U_dep_t sig A s)
         (u':U_dep_t sig A s')
         (pfu: Inn (A_dep_t sig A s) u)
         (pfu':Inn (A_dep_t sig A s') u')
         (pfe:s = s'),
    t = t' ->
    u = transfer_dep_r pfe u' ->
    Recursion_homo_aux a s t u pfu ->
    Recursion_homo_aux a s' t' u' pfu'.
  intros; subst;
  pose proof (proof_irrelevance _ pfu pfu');
  subst; auto.
Qed.


Lemma recursion_homo_aux_var_well_defined : 
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v)
         (s:S) (t:term X num_ops)
         (u u':U_dep_t sig A s)
         (pfu:Inn (A_dep_t sig A s) u)
         (pfu':Inn (A_dep_t sig A s) u'),
    Recursion_homo_aux_var a s t u pfu ->
    Recursion_homo_aux_var a s t u' pfu' ->
    u = u'.
  intros S X num_ops sig A v a s t u u' hu hu' h1 h2.
  inversion h1; subst.
  inversion h2; subst.
  apply existT_injective2 in H7; subst.
  apply existT_injective2 in H5; subst; auto.
Qed.


(*This monstrosity is a leftover from something that might or 
might not have worked out or work out, so I'll leave it for now, 
for aesthetics, but will make a case for its existence or (non-) 
in due time.*)
Lemma split_built_same_ind :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         {a:assment_dep (A_dep_t sig A) v}
         {R:forall      
               {op0} {pfop0:op0 < num_ops}
               {op1} {pfop1:op1 < num_ops}
               {t t':term X num_ops}
               {u:U_dep_t _ A (snd (sig op0 pfop0))}
               {u':U_dep_t _ A (snd (sig op1 pfop1))}
               {pfu:Inn (A_dep_t _ A (snd (sig op0 pfop0))) u}
               {pfu':Inn (A_dep_t _ A (snd (sig op1 pfop1))) u'},
               Recursion_homo_aux_op
                 a op0 pfop0 t u pfu ->  
               Recursion_homo_aux_op
                 a op1 pfop1 t' u' pfu' ->
               Prop}
         {op0} {pfop0:op0 < num_ops}
         {op1} {pfop1:op1 < num_ops},
    let n0 := projT1 (fst (sig op0 pfop0)) in
    let n1 := projT1 (fst (sig op1 pfop1)) in
    let w0 := projT2 (fst (sig op0 pfop0)) in
    let w1 := projT2 (fst (sig op1 pfop1)) in
    let s0 := snd (sig op0 pfop0) in
    let s1 := snd (sig op1 pfop1) in
    forall (ttm: assment_w_dep (Term_algebra_ens sig v) w0)
           (ttm':assment_w_dep (Term_algebra_ens sig v) w1)
           (ta:assment_w_dep (A_dep_t sig A) w0)
           (ta':assment_w_dep (A_dep_t sig A) w1)
           (pfen:n0 = n1),
      (forall j (pfj:j < n0),
         let pfj' := lt_eq_trans _ _ _ pfj pfen in
         let wj0 := nth_prod j pfj w0 in
         let wj1 := nth_prod j pfj' w1 in
         let jttm := nth_tuple_prod (proj1_sig ttm) j pfj in
         let jttm' := nth_tuple_prod (proj1_sig ttm') j pfj' in
         let jta := nth_tuple_prod (proj1_sig ta) j pfj in
         let jta' := nth_tuple_prod (proj1_sig ta') j pfj' in
         let pftaj :=
             iff1 (in_tuple_prod_ens_dep_nth_iff (A_dep_t sig A) _)
                  (proj2_sig ta) j pfj in
         let pftaj' :=
             iff1 (in_tuple_prod_ens_dep_nth_iff (A_dep_t sig A) _)
                  (proj2_sig ta') j pfj' in
         Recursion_homo_aux_var a wj0 jttm jta pftaj /\
         Recursion_homo_aux_var a wj1 jttm' jta' pftaj' \/
         (exists
             (pfjttm: jttm  <> nil)
             (pfjttm':jttm' <> nil)
             op2 (pfop2:op2 < num_ops)
             op3 (pfop3:op3 < num_ops)
             (pfe0:wj0 = snd (sig op2 pfop2))
             (pfe1:wj1 = snd (sig op3 pfop3)),
             let jta0 := transfer_dep pfe0 jta in
             let jta0' := transfer_dep pfe1 jta' in
             let pftaj0 := in_transfer_dep pfe0 _ pftaj in
             let pftaj0' := in_transfer_dep pfe1 _ pftaj' in
             exists
               (pfr2:Recursion_homo_aux_op
                       a op2 pfop2 jttm jta0 pftaj0)
               (pfr3:Recursion_homo_aux_op
                       a op3 pfop3 jttm' jta0' pftaj0'),
               hd' jttm pfjttm =
               inr (exist (fun c => c < num_ops) op2 pfop2) /\
               hd' jttm' pfjttm' =
               inr (exist (fun c => c < num_ops) op3 pfop3) /\
               R pfr2 pfr3)) ->
      (forall j (pfj:j < n0),
         let wj := nth_prod j pfj w0 in
         let jttm :=  nth_tuple_prod (proj1_sig ttm) j pfj in
         let jta := nth_tuple_prod (proj1_sig ta) j pfj in
         let pftaj :=
             iff1 (in_tuple_prod_ens_dep_nth_iff (A_dep_t sig A) _)
                  (proj2_sig ta) j pfj in
         Recursion_homo_aux_var a wj jttm jta pftaj \/
         (exists
             (pfjttm: jttm <> nil)
             op2 (pfop2:op2 < num_ops)
             (pfe:wj = snd (sig op2 pfop2)),
             let jta0 := transfer_dep pfe jta in
             let pftaj0 := in_transfer_dep pfe _ pftaj in
             hd' jttm pfjttm =
             inr (exist (fun c => c < num_ops) op2 pfop2) /\
             Recursion_homo_aux_op
               a op2 pfop2 jttm jta0 pftaj0)) /\
      (forall j (pfj:j < n1),
         let wj := nth_prod j pfj w1 in
         let jttm' :=  nth_tuple_prod (proj1_sig ttm') j pfj in
         let jta' := nth_tuple_prod (proj1_sig ta') j pfj in
         let pftaj' :=
             iff1 (in_tuple_prod_ens_dep_nth_iff (A_dep_t sig A) _)
                  (proj2_sig ta') j pfj in
         Recursion_homo_aux_var a wj jttm' jta' pftaj' \/
         (exists
             (pfjttm': jttm' <> nil)
             op3 (pfop3:op3 < num_ops)
             (pfe:wj = snd (sig op3 pfop3)),
             let jta0' := transfer_dep pfe jta' in
             let pftaj0' := in_transfer_dep pfe _ pftaj' in
             hd' jttm' pfjttm' =
             inr (exist (fun c => c < num_ops) op3 pfop3) /\
             Recursion_homo_aux_op
               a op3 pfop3 jttm' jta0' pftaj0')).
intros S X num_ops sig A v a R op0 hop0 op1 hop1; simpl;
intros ttm ttm' ta ta' hne h1; split; intros j hj.  
specialize (h1 j hj); destruct h1 as [h1 | h1]; destruct h1 as [h1 h2].  
left; auto. 
destruct h2 as [h2 [op2 [hop2 [op3 [hop3 [h3 [h4 [h5 [h6 [h7 [h8 h9]]]]]]]]]]]. 
right.
exists h1, op2, hop2, h3; split; auto.
specialize (h1 j (lt_eq_trans _ _ _ hj (eq_sym hne))).
destruct h1 as [h1|h1]; destruct h1 as [h1 h2].   
left.
term0 h2.
term1 h2.
term2 h2.
term3 h2.
gterm0.
pose proof (Recursion_homo_aux_var_functional
              a c2
              (nth_prod
                 j hj
                 (projT2
                    (fst (sig op1 hop1))))
              c1 (nth_tuple_prod (proj1_sig ttm') j hj) c0
              (nth_tuple_prod (proj1_sig ta') j hj) c c3) as h10.
unfold c, c0, c1, c2, c3 in h10.
hfirst h10.
apply nth_prod_functional3; auto.
specialize (h10 hf).  
hfirst h10.
f_equal; apply proof_irrelevance.
specialize (h10 hf0).
hfirst h10.
symmetry.
rewrite <- transfer_dep_r_eq_iff.
gtermr.
redterm0 y.
redterm0 c4.
fold c5.
pose proof (proof_irrelevance _ c5 hj) as h12.
rewrite h12; auto.  
specialize (h10 hf1 h2); auto.
destruct h2 as [h2 [op2 [hop2 [op3 [hop3 [h3 [h4 [h5 [h6 [h7 [h8 h9]]]]]]]]]]].
right.
ex_goal.
nterml h2.
redterm0 x.
pose proof (proof_irrelevance _ c hj) as h12.
rewrite <- h12; auto.
exists hex.
exists op3, hop3.
ex_goal.
rewrite <- h4.
f_equal; apply proof_irrelevance.
exists hex0; split; auto.
rewrite <- h8.
apply f_equal_dep.
repeat f_equal; apply proof_irrelevance.
                        
pose proof h6 as h6'.
gen0.
intro h13.

revert h9; revert h6. 
Gen0.
intros h14 h15 h16. 
term0 h13; term0 h14.
redterm0 c0.
redterm0 c1.
assert (h17:c = c0).
  unfold c, c0.
  rewrite transfer_dep_hetero.
  pose proof (nth_tuple_prod_functional3 (proj1_sig ta') j hj c2) as h18.
  rewrite h18.
  fold c2.
  f_equal; apply proof_irrelevance.
symmetry in h17.
pose proof (subsetT_eq_compat _ _ _ _ h14 h13 h17) as h18.
dependent rewrite <- h18. 
pose proof h15 as h15'.
fold c0 c c2 in h15'.
fold c0.
apply (recursion_homo_aux_op_functional
              a op3 op3 hop3 hop3
              (nth_tuple_prod (proj1_sig ttm') j c2)
              (nth_tuple_prod (proj1_sig ttm') j hj)
              c0 c0 h14 h14 (eq_refl _)).
f_equal; apply proof_irrelevance.
rewrite transfer_seg_fun_depT2_r_eq_refl; auto.
assumption.
Qed.


(*Turns out "subterm injectivity" is non-trivial, though 
still very manageable, and rather than 
postpone a Winter release, I thought I 
would just include the last step in [recursion_homo_ex]
in a bit of a clumsy way -- by leaving it as an Axiom, more or less 
an Admitted, and the proof is in the works, as you can see by all 
the nesting functions and lemmas, and an easy induction on depth 
should cinch this, and something a bit stronger that will be reworked 
in [recursion_homo_ex].
Release date over purity of release is what it's about, so long as 
the releases come yearly at Winter Solstice!*)
Axiom terms_app_list_eq_length : 
  forall {S X num_ops n} {sig:signature S num_ops}
         (v:X->S)
         (p:S^n)
         (t u:term X num_ops^n),
    (forall i (pfi:i < n),
       Inn (Term_algebra_ens
              sig v
              (nth_prod i pfi p))
           (nth_prod i pfi t)) ->
    (forall i (pfi:i < n),
       Inn (Term_algebra_ens
              sig v
              (nth_prod i pfi p))
           (nth_prod i pfi u)) ->
    (app_list (nprod_to_list _ _ t) =
     app_list (nprod_to_list _ _ u)) ->
    forall i (pfi:i < n),
      length (nth_prod i pfi t) =
      length (nth_prod i pfi u).   


Definition recursion_homo_ex :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v)
         (s:S)
         (t:term X num_ops)
         (pft:Inn (Term_algebra_ens sig v s) t),
  exists! u:U_dep_t sig A s,
    exists pfu:Inn (A_dep_t sig A s) u,
      Recursion_homo_aux a s t u pfu.
  intros S X num_ops sig A v a s t ht.   
  revert a; revert A.
  induction ht.
  intros A a.
  exists (proj1_sig (a x)).
  red; split.
  exists (proj2_sig _).
  constructor.
  intros u hu.
  destruct hu as [hu h1].
  inversion h1; subst.
  clear h1.
  apply existT_injective2 in H2; auto.
  inversion H2.
  intros A a. 
  exists (proj1_sig
            (ops_dep_t
               sig A i pfi  
               (exist
                  (fun c =>
                     Inn (tuple_prod_ens_dep (A_dep_t sig A)
                                             (projT2 (fst (sig i pfi)))
                         )
                         c)
                  _
                  (in_transfer_tt_tuple_prod_ens_dep
                     (U_dep_t sig A)
                     (A_dep_t sig A)
                      (projT2 (fst (sig i pfi)))
                      H)))). 
  red.
  split.
  exists (proj2_sig _). 
  constructor.
  constructor.  
  intros u hu.
  destruct hu as [h1 h2].  
  inversion h2; clear h2; subst.
  inversion H3; clear H3 H9.
  unfold u1, pfintt0, w, n, s, w0 in H2.
  clear u1 pfintt0 w n s w0.
  subst.
  pose proof (proof_irrelevance _ pfop pfi); subst.
  pose proof (proof_irrelevance _ H pfeq0); subst.
  apply existT_injective2 in H6; auto.
  clear H10.
  unfold opta, w, n, s, w0, ttm'0, w0
    in H7, ttm, ta0, ttm', ttm, ttm0, tm0, H8.
  clear opta w n s w0 ttm'0 tm0 ttm'.
  subst.
  pose proof (proof_irrelevance _ pfi pfop); subst.
  apply existT_injective2 in H6.
  rewrite <- H6.
  repeat f_equal.
  apply proj1_sig_injective; simpl.
  apply tuple_prod_n0_eq; auto.

  intros A a. 
  pose (fun j (pfj:j < projT1 (fst (sig i pfi))) =>
          (proj1_sig
             (constructive_definite_description
                _
                (H0 j pfj A a)))) as f. 
  pose proof (rec_homo_aux_op_app a i pfi) as h3. 
  simpl in h3.
  pose (seg_fun_prod_ens_to_assment
          (Term_algebra_ens sig v)
          (projT2 (fst (sig i pfi)))
          _
          H) as tm_w.
  specialize (h3 tm_w).
  pose (seg_fun_prod_ens_to_assment
          (A_dep_t sig A)
          (projT2 (fst (sig i pfi)))
          f) as ta.
  hfirst ta.
  intros j hj. 
  unfold f.
  destruct constructive_definite_description; simpl. 
  destruct e; auto.
  pose (ta hf) as ta'.
  specialize (h3 ta').    
  hfirst h3.
  unfold ta', ta, tm_w.
  intros j hj.
  unfold f.
  simpl.
  match goal with
      |- Recursion_homo_aux_var _ _ _ _ ?pf' \/ _ => generalize pf' end.
  do 2 rewrite nth_tuple_prod_seg_fun_depT_nth_to_tuple_prod. 
  intro h9.
  match goal with
      |- Recursion_homo_aux_var _ _ _ ?c' _ \/ _ => pose c' as c3 end.
  redterm0 c3.  
  pose proof (proj2_sig c) as h10.
  simpl in h10.
  destruct h10 as [h10 h11].
  
  inversion h11; clear h11.
  left.
    gen0.
  unfold c in H7.
  dependent rewrite <- H7.
  rewrite <- H3.
  intro.
  pose proof (proof_irrelevance _ h0 (proj2_sig (a x))) as h12.
  rewrite h12.
  constructor.
  right.
  subst.
  unfold c in H7.
  pose proof (recursion_homo_aux_op_hd _ _ _ _ _ _ H4) as h8.
  destruct h8 as [h8 h13].
  exists h8, op, pfop, (eq_sym H6). 
  split; auto.
  pose proof H7 as h14.
  symmetry in h14.
  rewrite (transfer_dep_eq_iff _ _ (eq_sym H6)) in h14.
  terml h14.
  pose proof (recursion_homo_aux_op_functional
                a op op pfop pfop
                (nth_prod j hj t)
                (nth_prod j hj t)
                u x
                pfu) as h15.
  gterm0.
  fold c0.
  specialize (h15 c0 (eq_refl _) (eq_refl _)).
  hfirst h15.
  unfold x.
  rewrite h14, transfer_seg_fun_depT2_r_eq_refl; auto.
  specialize (h15 hf0).
  unfold x in h15.
  apply h15; auto.
  specialize (h3 hf0). 
  term1 h3.
  exists c.
  red; split.
  exists (proj2_sig _).
  constructor.  
  unfold c.
  unfold tm_w in h3.
  rewrite op_term_to_term_sig_tuple_prod_ens_dep_const_to_term_algebra in h3; auto.
  intros u h1.
  destruct h1 as [h1 h4].
  unfold c. 
  inversion h4; clear h4.
  subst.
  inversion H4; clear H4.
  clear n s.
  unfold w0, pfintt0 in u1.
  unfold u1, w0, pfintt0, w0 in H3.
  unfold w0 in u1.
  unfold c.
  clear H11 w0 pfintt0 u1 w c. 
  subst.
  pose proof (proof_irrelevance _ pfi pfop).
  clear H7.
  unfold f, ta', ta, f.
  unfold f, ta', ta, f, tm_w in hf, h3, hf0.
  clear f ta' ta tm_w.
  subst. 
  apply existT_injective2 in H5; subst.
  repeat f_equal.
  apply proj1_sig_injective; simpl.
  rewrite seg_fun_depT_nth_to_tuple_prod_eq_iff.
  intros k hk.
  destruct constructive_definite_description; simpl.
  destruct e as [h10 h11].  
  inversion h11; omega; clear h11.
  unfold n in H10.
  clear n.
  unfold w.
  unfold w in ttm.
  clear ttm' w.
  unfold opta in H8, H11.
  clear opta.
  unfold w0 in ta1, ttm0, H10.
  unfold ttm'0, w0 in tm0, H9.
  clear ttm'0.
  unfold w0 in tm0, H8, H11.
  clear w0.
  subst.
  unfold ta', ta in h3, hf0.
  unfold ta', ta.
  clear c.
  clear ta' ta.
  unfold op_term_to_term in H9.
  pose proof (proof_irrelevance _ pfi pfop).
  unfold f.
  unfold f, tm_w in hf, h3, hf0.
  clear f tm_w.
  subst.
  apply existT_injective2 in H7.
  subst.
  repeat f_equal.  
  apply proj1_sig_injective; simpl.
  rewrite seg_fun_depT_nth_to_tuple_prod_eq_iff. 
  intros k hk.
  destruct constructive_definite_description; simpl.
  destruct e as [e he].
  pose proof (H0 k hk A a) as h12.
  destruct h12 as [u h12].
  red in h12.
  destruct h12 as [[h12 h13] h14].
  symmetry.
  pose proof (h14 x) as h15.
  hfirst h15.
  exists e; auto.
  specialize (h15 hf1).
  subst.
  symmetry.
  apply h14.
  exists (in_nth_tuple_prod_ens_dep (proj2_sig ta1) k hk).  
  apply app_list_nprod_to_list_inj in H9; subst.
  pose proof (nth_prod_sig_tuple_prod_ens_dep_const_to
                (Term_algebra_ens sig v)
                (projT2 (fst (sig i pfop))) ttm0 k hk) as h15.
  terml h15.
  termr h15.
  gterm0.
  symmetry in h15.
  pose proof (Recursion_homo_aux_functional
                a
                (nth_prod k hk (projT2 (fst (sig i pfop))))
                (nth_prod k hk (projT2 (fst (sig i pfop))))
                y x0
                (nth_tuple_prod (proj1_sig ta1) k hk)
                (nth_tuple_prod (proj1_sig ta1) k hk)
                c c (eq_refl _)) as h16.
  unfold x0, y in h16.
  specialize (h16 h15 (eq_refl _)).
  unfold c in h16.
  apply h16; clear h16.
  pose proof (H10 k hk) as h17.
  destruct h17 as [h17|h17].
  inversion h17; clear h17.
  rewrite (transfer_dep_eq_iff
             _ _ H12) in H13.
  pose proof (Recursion_homo_aux_functional
                a
                (v x1)
                (nth_prod k hk (projT2 (fst (sig i pfop))))
                (inl x1::nil)
                (inl x1::nil)
                (proj1_sig (a x1))
                (nth_tuple_prod (proj1_sig ta1) k hk)
                (proj2_sig (a x1))
                (in_nth_tuple_prod_ens_dep (proj2_sig ta1) k hk)
                H4
                (eq_refl _)) as h18.
  hfirst h18.
  symmetry.
  rewrite <- transfer_dep_r_eq_iff; auto.
  specialize (h18 hf2).
  apply h18.
  constructor.
  destruct h17 as [hn [op [hop [h19 [h20 h21]]]]].
  term0 h21.
  eapply (Recursion_homo_aux_functional
           a
           (snd (sig op hop))
           (nth_prod k hk (projT2 (fst (sig i pfop))))
           (nth_tuple_prod (proj1_sig ttm0) k hk)
           (nth_tuple_prod (proj1_sig ttm0) k hk)
           (transfer_dep h19 (nth_tuple_prod (proj1_sig ta1) k hk))); auto.
  erewrite transfer_dep_transfer_dep_r_compat.
  reflexivity.
  constructor.
  apply h21.
  eapply terms_app_list_eq_length; auto.
  intros j hj.
  rewrite (nth_prod_sig_tuple_prod_ens_dep_const_to _ _ ttm0).
  pose proof (proj2_sig ttm0) as h15.
  simpl in h15.
  destruct h15 as [h15]; auto.
  Grab Existential Variables.
  auto.
Qed.


Lemma recursion_homo_aux_well_defined :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v)
         (s:S)
         (t:term X num_ops)
         (pft:Inn (Term_algebra_ens sig v s) t)
         u u'
         (pfu:Inn (A_dep_t sig A s) u)
         (pfu':Inn (A_dep_t sig A s) u'),
    Recursion_homo_aux a s t u pfu ->
    Recursion_homo_aux a s t u' pfu' ->
    u = u'.
  intros S X num_ops sig A v a s t ht u u' hu hu' h1 h2.
  pose proof (recursion_homo_ex a s t ht) as h3.
  destruct h3 as [c [h4 h5]].
  destruct h4 as [h4 h6].
  pose proof (h5 u) as h7.
  hfirst h7.
  exists hu; auto.
  specialize (h7 hf); subst.
  apply h5.
  exists hu'; auto.
Qed.


Lemma recursion_homo_aux_op_well_defined :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v)
         op (pfop:op < num_ops)
         (t:term X num_ops)
         (pft:Inn (Term_algebra_ens sig v (snd (sig op pfop))) t)
         u u'
         (pfu:Inn (A_dep_t sig A (snd (sig op pfop))) u)
         (pfu':Inn (A_dep_t sig A (snd (sig op pfop))) u'),
    Recursion_homo_aux_op a op pfop t u pfu ->
    Recursion_homo_aux_op a op pfop t u' pfu' ->
    u = u'.
  intros S X num_ops sig A v a op pfop t ht u u' hu hu' h1 h2.
  pose proof (recursion_homo_aux_op a op pfop t u hu h1) as h3.
  pose proof (recursion_homo_aux_op a op pfop t u' hu' h2) as h4.
  pose proof (recursion_homo_aux_well_defined _ _ _ ht _ _ _ _ h3 h4); auto.
Qed.




Definition recursion_homo 
           {S X num_ops}
           {sig:signature S num_ops}
           {A:TAlgebra_dep sig} {v:X->S}
           (a:assment_dep (A_dep_t sig A) v)
           (s:S)
           (t:term X num_ops)
           (pft:Inn (Term_algebra_ens sig v s) t) :
  U_dep_t sig A s :=
  proj1_sig (constructive_definite_description
               _
               (recursion_homo_ex a s t pft)).

Lemma recursion_homo_compat :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v)
         (s:S)
         (t:term X num_ops)
         (pft:Inn (Term_algebra_ens sig v s) t),
  exists pfu:Inn (A_dep_t sig A s) (recursion_homo a s t pft),
    Recursion_homo_aux a s t (recursion_homo a s t pft) pfu.
  unfold recursion_homo; intros;
  destruct constructive_definite_description; auto.
Qed.

Lemma in_recursion_homo :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v)
         (s:S)
         (t:term X num_ops)
         (pft:Inn (Term_algebra_ens sig v s) t),
    Inn (A_dep_t _ A s)
        (recursion_homo a s t pft).
  intros S X num_ops sig A v a s t ht.
  pose proof (recursion_homo_compat a s t ht) as h1.
  destruct h1; auto.
Qed.



Lemma recursion_homo_eq_iff :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v)
         (s:S)
         (t:term X num_ops)
         (pft:Inn (Term_algebra_ens sig v s) t)
         (u:U_dep_t _ A s),
    recursion_homo a s t pft = u <->
    exists (pfu:Inn (A_dep_t _ A s) u),
      Recursion_homo_aux a s t u pfu.
  intros S X num_ops sig A v a s t ht u.
  pose proof (recursion_homo_compat a s t ht) as h1.
  destruct h1 as [h1 h2].
  split; intro h3; subst.
  exists h1; auto.
  destruct h3 as [h3 h4].
  eapply recursion_homo_aux_well_defined.
  apply ht.
  apply h2.
  apply h4.
Qed.



Lemma recursion_homo_vx :
  forall {S X num_ops} {sig:signature S num_ops} {A:TAlgebra_dep sig}
         {v:X->S} (a:assment_dep (A_dep_t sig A) v)
         x
         (pfin:Inn (Term_algebra_ens sig v (v x)) (inl x :: nil)),
    recursion_homo a (v x) (inl x :: nil) pfin =
    proj1_sig (a x).
  intros S X num_ops sig A v a x h1.
  pose proof (recursion_homo_compat a (v x) (inl x ::nil) h1) as h2.
  destruct h2 as [h2 h3].
  inversion h3; clear h3; subst.
  clear H3.
  apply existT_injective2 in H2; auto.
  inversion H2.
Qed.


Lemma recursion_homo_nil_op :
  forall
    {S X num_ops}
    {sig:signature S num_ops}
    {v:X->S}
    {op} {pfop:op < num_ops}
    (A:TAlgebra_dep sig)
    (a:assment_dep (A_dep_t sig A) v)
    s
    (pfe:projT1 (fst (sig op pfop)) = 0)
    (pfint:Inn (Term_algebra_ens sig v s)
               (inr (exist (fun c3 : nat => c3 < num_ops) op pfop) :: nil))
    (pfe':snd (sig op pfop) = s),
    recursion_homo
      a s
      (inr
         (exist (fun c3 : nat => c3 < num_ops) op pfop) :: nil) pfint =
    transfer_dep
      pfe'
      (proj1_sig (ops_dep_t
                    sig A op pfop
                    (exist
                       (fun c =>
                          Inn (tuple_prod_ens_dep (A_dep_t sig A)
                                                  (projT2 (fst (sig op pfop))))
                              c)
                       _ (in_transfer_tt_tuple_prod_ens_dep
                            (U_dep_t sig A)
                            (A_dep_t sig A)
                            (projT2 (fst (sig op pfop)))
                            pfe)))).
  intros S X num_ops sig v op hop A a x h1 h2 h3; subst.
  rewrite transfer_dep_eq_refl.
  pose proof
       (recursion_homo_compat
          a (snd (sig op hop)) (inr (exist (fun c => c < num_ops) op hop) :: nil) h2) as h4. 
  destruct h4 as [h4 h5].
  inversion h5; clear h5; subst.
  inversion H2; clear H2.
  clear s n.
  unfold u0.
  unfold u0, pfintt0, w0, w in H1, H8.
  clear u0 pfintt0 w0 w. 
  subst.
  pose proof (proof_irrelevance _ pfop hop); subst.
  pose proof (proof_irrelevance _ h1 pfeq0); subst.
  rewrite transfer_dep_eq_iff in H5.
  rewrite <- H5.
  rewrite transfer_dep_eq_refl.
  reflexivity.
  clear tm0.
  unfold ttm'0 in H7.
  clear ttm'0 s ttm'.
  unfold opta, w0, w, n in H6, H9, ttm, ta0, ttm0, H8, H7.
  clear opta w0 w n.
  subst.
  rewrite transfer_dep_eq_iff in H5.
  rewrite <- H5.
  pose proof (proof_irrelevance _ hop pfop); subst.
  pose proof (@tuple_prod_n0_eq
                _ _ _ _
                (proj1_sig ta0)
                (transfer_tt_tuple_prod (U_dep_t sig A) h1
                                        (projT2 (fst (sig op pfop)))) h1) as h5.
  generalize (proj2_sig ta0).
  generalize h1.
  rewrite h5.
  intros.
  rewrite transfer_dep_eq_refl.
  repeat f_equal.
  apply proj1_sig_injective; simpl; auto.
  f_equal.
  apply proof_irrelevance.

  Grab Existential Variables.
  repeat f_equal.
  apply proof_irrelevance.
  reflexivity.
Qed.


Definition recursion_homo_Sfun_dep
           {S X num_ops}
           {sig:signature S num_ops}
           {A:TAlgebra_dep sig} {v:X->S}
           (a:assment_dep (A_dep_t sig A) v) :
  Sfun_dep (Term_algebra_ens sig v)
           (A_dep_t sig A) :=
  fun s
      (t:term X num_ops)
      (pft:Inn (Term_algebra_ens sig v s) t) =>
    let (pfu, _) :=
        constructive_definite_proof
          (recursion_homo_compat a s t pft) in
    exist (fun c => Inn (A_dep_t sig A s) c)
          (recursion_homo a s t pft)
          pfu.



Lemma recursion_extends :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v)
         (x:X),
    recursion_homo_Sfun_dep
      a (v x) (inl x::nil)
      (term_algebra_ensX sig v x) =
    a x.
  intros S X num_ops sig A v a.
  unfold recursion_homo_Sfun_dep.
  intro x.
  unfold constructive_definite_proof; simpl.
  pose proof
       (recursion_homo_compat a (v x) (inl x::nil) (term_algebra_ensX sig v x)) as h1.
  destruct h1 as [h1 h2].
  pose proof (recursion_homo_aux_var a x) as h3.
  apply proj1_sig_injective; simpl.
  apply (recursion_homo_aux_well_defined
           _ _ _ (term_algebra_ensX _ _ _) _ _ _ _  h2 h3).
Qed.



Lemma homo_recursion :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig} {v:X->S}
         (a:assment_dep (A_dep_t sig A) v),
    Thomo_dep (recursion_homo_Sfun_dep a)
              (A := Term_algebra_dep (type_eq_decS_t _ A) sig v) (B := A).
  intros S X num_ops sig A v a.  
  constructor. 
  intros i hi; simpl.
  intros a'.
  destruct a' as [p hp].  
  pose proof (recursion_homo_compat a _ _ (proj2_sig (Term_algebra_dep' sig v i hi (exist _ _ hp)))) as h1.  
  destruct h1 as [h1 h2].
  unfold recursion_homo_Sfun_dep.
  unfold constructive_definite_proof.
  destruct constructive_definite_description.
  apply proj1_sig_injective; simpl.
  pose proof (proof_irrelevance _ x h1); subst.
  rename r into h10.
  inversion h10; clear h10; subst.
  inversion H2; clear H2.
  clear H9.
  unfold u0 in H1; clear H1 u0 s n pfintt0 w0 w.
  subst.
  unfold Term_algebra_op_aux_dep, op_term_to_app_term.
  termr H7. 
  redterm0 y.
  redterm0 c.
  gterml. 
  redterm1 x. 
  redterm0 c1.
  redterm0 c2.
  redterm0 c3.
  pose proof (proof_irrelevance _ c4 c0) as h8.
  gen0.
  intro h10; simpl in h10.
  pose proof h10 as h10'.
  unfold op_term_to_app_term in h10'.
  rewrite <- H7 in h10'.
  pose proof (f_equal (fun c => inr (exist (fun c => c < num_ops) i hi)::c) H7) as h12.
  pose proof (subsetT_eq_compat _ _ _ _ h10' h10 h12) as h13.
  simpl in h13; simpl.
  unfold op_term_to_app_term in h13.
  dependent rewrite <- h13.
  pose proof (proof_irrelevance _ pfop hi); subst.
  erewrite recursion_homo_nil_op.
  rewrite transfer_dep_eq_refl.
  repeat f_equal.
  apply proj1_sig_injective; simpl.
  apply tuple_prod_n0_eq; auto.
  clear tm0.
  unfold opta, ttm'0, w0, w, n in H7, ta0, ttm0, H8, H6, H9, ttm.
  clear opta ttm'0 s ttm' w0 w n.
  subst.
  pose proof (proof_irrelevance _ pfop hi); subst.
  apply existT_injective2 in H5.
  rewrite <- H5.
  repeat f_equal.
  apply proj1_sig_injective; simpl.
  symmetry. 
  rewrite proj1_sig_map_assment_w_dep_eq_iff.
  intros k hk; simpl.
  rewrite recursion_homo_eq_iff.
  exists (in_nth_tuple_prod_ens_dep (proj2_sig ta0) k hk).
  rename H7 into h12.
  unfold op_term_to_term in h12.
  apply app_list_nprod_to_list_inj in h12; simpl in h12.
  unfold sig_tuple_prod_ens_dep_const_to in h12; simpl in h12.
  destruct ttm0 as [q hq]; simpl in h12.
  apply inj_transfer_tuple_prod_const in h12; subst.
  specialize (H8 k hk); simpl in H8.
  destruct H8 as [h10 |h10].
  inversion h10; clear h10.
  pose proof (recursion_homo_aux_var a x) as h11.
  gen0.
  dependent rewrite <- H11.
  intro h12.
  pose proof (proof_irrelevance _ h12 (proj2_sig (a _))); subst.
  constructor.
  destruct h10 as [h10 [op [hop [h13 h14]]]].
  destruct h14 as [h14 h15].
  pose proof (recursion_homo_aux_op
                a op hop (nth_tuple_prod p k hk)
                (transfer_dep h13 (nth_tuple_prod (proj1_sig ta0) k hk)) _
                h15) as h16.
  pose proof (Recursion_homo_aux_functional
                a (snd (sig op hop))
                (nth_prod k hk (projT2 (fst (sig i hi))))
                (nth_tuple_prod p k hk)
                (nth_tuple_prod p k hk)
                (transfer_dep h13 (nth_tuple_prod (proj1_sig ta0) k hk))
                (nth_tuple_prod (proj1_sig ta0) k hk)) as h17.
  eapply h17; auto. 
  rewrite (transfer_dep_transfer_dep_r_compat h13 (eq_sym h13)).
  reflexivity. 
  apply h16.
  intros; simpl.
  eapply terms_app_list_eq_length.
  intros j hj.   
  pose proof (proj2_sig ttm0) as h15.
  simpl in h15.
  destruct h15 as [h15].
  specialize (h15 j hj).
  rewrite <- (nth_prod_sig_tuple_prod_ens_dep_const_to _ _ ttm0) in h15.
  apply h15.
  intros j hj.
  pose proof hp as h16.
  destruct h16 as [h16].
  rewrite nth_prod_transfer_tuple_prod_const; auto.
  auto.

  Grab Existential Variables.
  reflexivity.
  auto.
Qed.



Lemma transfer_dep_nth_tuple_prod_proj1_map_assment_w_dep :
  forall {S X num_ops}
         {sig:signature S num_ops}
         {A:TAlgebra_dep sig}
         {v:X->S}
         {op} {pfop:op < num_ops} 
         (a:assment_dep (A_dep_t sig A) v)
         (a':assment_w_dep
               (Term_algebra_ens sig v)
               (projT2 (fst (sig op pfop))))
         j
         (hj:j < projT1 (fst (sig op pfop)))
         x
         (pfe:nth_prod j hj (projT2 (fst (sig op pfop))) = v x),
    transfer_dep pfe
             (nth_tuple_prod 
                (proj1_sig
                   (map_assment_w_dep
                      (recursion_homo_Sfun_dep a)
                      a'))
                j hj) =
    recursion_homo a (v x) (nth_tuple_prod (proj1_sig a') j hj)
                   (in_nth_tuple_prod_term_algebra a' j hj x pfe).
  unfold recursion_homo_Sfun_dep; intros.    
  rewrite nth_tuple_prod_map_assment_w_dep.
  unfold constructive_definite_proof; simpl.  
  rewrite <- transfer_dep_eq_iff.
  gterml.
  gtermr.
  redterm0 x0.
  redterm0 c.
  redterm0 y.
  redterm0 c1.
  fold c2 c0. 
  generalize c0, c2; clear c0 c2.
  clear c1 c y x0.
  rewrite pfe.
  intros; repeat f_equal.
  apply proof_irrelevance.
Qed.


Lemma transfer_dep_nth_tuple_prod_proj1_map_assment_w_dep' : 
  forall {S X num_ops} {sig:signature S num_ops} {A:TAlgebra_dep sig}
         {v:X->S}
         {op}
         {pfop:op < num_ops}
         (a:assment_dep (A_dep_t sig A) v)
         (a':assment_w_dep
               (Term_algebra_ens sig v)
               (projT2 (fst (sig op pfop))))
         x
         j
         (pfj:j < projT1 (fst (sig op pfop)))
         (pfe:nth_prod j pfj (projT2 (fst (sig op pfop))) = v x),
    transfer_dep
      pfe
      (nth_tuple_prod
         (proj1_sig
            (map_assment_w_dep
               (fun (s : S) (t : term X num_ops)
                    (pft : Inn (Term_algebra_ens sig v s) t) =>
                  exist
                    (fun c1 : U_dep_t sig A s => Inn (A_dep_t sig A s) c1)
                     (recursion_homo a s t pft)
                     (proj1_sig
                        (constructive_definite_description
                           (Recursion_homo_aux a s t
                                               (recursion_homo a s t pft))
                           (propositional_ex_unq
                              (Inn (A_dep_t sig A s)
                                   (recursion_homo a s t pft))
                              (Recursion_homo_aux a s t
                                                  (recursion_homo a s t pft))
                              (recursion_homo_compat a s t pft))))) a'))
         j pfj) = 
    recursion_homo
      a (v x) (nth_tuple_prod (proj1_sig a') j pfj)
      (in_nth_tuple_prod_term_algebra a' j pfj x pfe).
  intros;
  rewrite <- transfer_dep_nth_tuple_prod_proj1_map_assment_w_dep; auto.
Qed.



Lemma recursion_homo_aux_pred_op :
  forall {S X num_ops} {sig:signature S num_ops} {A:TAlgebra_dep sig}
         {v:X->S}
         (a:assment_dep (A_dep_t sig A) v)
         i (pfi:i < num_ops)
         (a':assment_w_dep
               (Term_algebra_ens sig v)
               (projT2 (fst (sig i pfi))))
         (pfr:Inn (A_dep_t sig A (snd (sig i pfi)))
                  (recursion_homo a (snd (sig i pfi))
                                  (proj1_sig (Term_algebra_dep' sig v i pfi a'))
                                  (proj2_sig (Term_algebra_dep' sig v i pfi a')))),
    Recursion_homo_aux a (snd (sig i pfi))
                     (proj1_sig (Term_algebra_dep' sig v i pfi a'))
                     (recursion_homo a (snd (sig i pfi))
                                     (proj1_sig (Term_algebra_dep' sig v i pfi a'))
                                     (proj2_sig (Term_algebra_dep' sig v i pfi a'))) pfr <->
    forall j (pfj:j < projT1 (fst (sig i pfi))),
      (exists x:X,
         v x = nth_prod j pfj (projT2 (fst (sig i pfi)))) \/
      (exists op (pfop:op < num_ops),
         snd (sig op pfop) = nth_prod j pfj (projT2 (fst (sig i pfi)))).
  intros S X num_ops sig A v a i hi a' hr.
  split; intro h1.
  inversion h1; clear h1. 
  intros j hj.
  left.
  exists x.
  unfold Term_algebra_dep' in H.
  destruct a' as [y z].
  simpl in H.
  unfold Term_algebra_op_aux_dep, op_term_to_app_term in H.
  discriminate.
  intros j hj.
  unfold Term_algebra_dep' in H5; simpl in H5, H3.
  unfold Term_algebra_op_aux_dep, op_term_to_app_term in H5, H3, H2, hr. 
  destruct a' as [p hp].
  subst.
  simpl in H5, H3, H2, hr.
  clear H4 H3.
  inversion H2; clear H2.
  clear s n.
  unfold u0, pfintt0, w0, w in H1, H7.
  clear u0 pfintt0 w0 w.
  subst.
  pose proof (proof_irrelevance _ hi pfop); subst.
  clear H0.
  right.
  pose proof hj as hj'.
  rewrite pfeq in hj'.
  contradict (lt_n_0 _ hj').
  clear H7.
  unfold opta, w0, ttm'0, tm0, n in H3, H4, ta0, ttm0, H6, H4.
  clear opta w0 ttm'0 tm0 s ttm' n.
  unfold w in ttm.
  clear w.
  subst.
  pose proof (proof_irrelevance _ hi pfop); subst.
  specialize (H6 j hj).
  destruct H6 as [h5 | h5].
  inversion h5; clear h5.
  left.
  exists x; auto.
  destruct h5 as [h5 [op [hop [h8 h9]]]].
  right.
  exists op, hop; auto.
  pose proof (recursion_homo_compat
                a (snd (sig i hi))
                (proj1_sig (Term_algebra_dep' sig v i hi a'))
                (proj2_sig _)) as h2.
  destruct h2.
  pose proof (proof_irrelevance _ x hr);subst; auto.
Qed.



Theorem RecursionTheorem :
  forall {S X num_ops} {sig:signature S num_ops} {A:TAlgebra_dep sig}
         {v:X->S} (a:assment_dep (A_dep_t sig A) v),
  exists! (f:Sfun_dep (Term_algebra_ens sig v) (A_dep_t sig A)),
    Thomo_dep f
              (A := Term_algebra_dep (type_eq_decS_t _ A) sig v)
              (B := A) /\
    forall x:X,
      let pfx := term_algebra_ensX sig v x in
      let fx := proj1_sig (f _ _ pfx) in
      let ax := proj1_sig (a x) in
      fx = ax.
intros S X num_ops sig A v a.  
exists (recursion_homo_Sfun_dep a).
red; split.
split.
apply homo_recursion.
intro; f_equal; apply recursion_extends.
intros f h1.
destruct h1 as [h1 h2].
eapply (homos_eq_on_generators (A := Term_algebra_dep (type_eq_decS_t _ A) sig v) (B := A)); auto. 
apply homo_recursion.
eapply pre_sys_generates_term_algebra'.
intros s u hin.
destruct hin as [x]; subst.
destruct i as [i]; subst.
erewrite f_equal_dep.
rewrite recursion_extends.
rewrite <- h2; repeat f_equal; apply proof_irrelevance.
reflexivity.
intros op hop n w s he; simpl.
destruct h1 as [h1].
specialize (h1 op hop); simpl in h1.
pose proof (in_transfer_tt_tuple_prod_ens_dep
                        (U_dep_t sig A)
                        (A_dep_t sig A)
                        w he) as h3.

pose proof (in_transfer_tt_tuple_prod_ens_dep
              (fun _ => term X num_ops)
              (Term_algebra_ens sig v) w he) as h4. 
specialize (h1 (exist
                  (fun c =>
                     Inn (tuple_prod_ens_dep
                            (Term_algebra_ens sig v) w) c) _ h4)).
unfold s, n, w; unfold w in h4, h1, h3; clear s n w. 
unfold Term_algebra_dep' in h1; simpl in h1.
unfold recursion_homo_Sfun_dep.
unfold constructive_definite_proof.
destruct constructive_definite_description; simpl.
apply proj1_sig_injective; simpl. 
pose proof (recursion_homo_nil_op A a (snd (sig op hop)) he) as h12.
unfold Term_algebra_op_aux_dep, op_term_to_app_term, op_term_to_term; simpl.
gterml.
redterm1 x0.
redterm0 c.
assert (h8:c0 = nil).
  unfold c0.
  rewrite app_list_eq_nil_iff.
  left.
  gterml.
  redterm0 x1.
  pose proof (nprod_to_list_eq_nil_iff c1) as h9.
  fold c1.
  rewrite h9 at 1; auto.
fold c0.
pose proof (f_equal (cons (inr (exist (fun q => q < num_ops) op hop))) h8)
  as h9.
redterm0 x0.
fold c1.
pose proof c1 as hc1; simpl in hc1.
unfold Term_algebra_op_aux_dep, op_term_to_app_term, op_term_to_term in hc1, c1; simpl in hc1, c1.
fold c0 in hc1.
rewrite h9 in hc1.
pose proof (subsetT_eq_compat _ _ _ _ c1 hc1 h9) as h10.
fold c0 c1 in h10.
dependent rewrite -> h10.
terml h1.
redterm0 x1.
specialize (h12 hc1 (eq_refl _)). 
rewrite h12 at 1. 
rewrite transfer_dep_eq_refl.

unfold c2, Term_algebra_op_aux_dep, op_term_to_app_term, op_term_to_term
  in h1, c2;
  simpl in h1, c2.
fold c0 c2 in h1.
pose proof (proof_irrelevance _ c1 c2) as h20.
rewrite <- h20 in h1.
dependent rewrite -> h10 in h1.
rewrite h1.
repeat f_equal.
apply proj1_sig_injective; simpl.
apply tuple_prod_n0_eq; auto.
Qed.



(*Existence equation.*)
Inductive ex_eq {S X num_ops} (v:X->S) (t t':term X num_ops) : Prop :=
  ex_eq_intro :  forall (sig:signature S num_ops) s,
                     Inn (Term_algebra_ens sig v s) t  ->
                     Inn (Term_algebra_ens sig v s) t' ->
                     ex_eq v t t'.

Definition ex_eq_ens {S X} num_ops (v:X->S) :
  Type :=
  Ensemble {pr:(term X num_ops) * (term X num_ops) |
            ex_eq v (fst pr) (snd pr)}.


Definition dom_cond {S X} num_ops (v:X->S) :
  Type :=
  ex_eq_ens num_ops v.

(*Equational partial signature.*)
Definition ep_signature {X S} (v:X->S) num_ops : Type :=
  seg_fun num_ops ((all_tuples S * S) * (dom_cond num_ops v))%type.

Definition ep_signature_to_sig 
           {X S} {v:X->S} {num_ops}
           (theta:ep_signature v num_ops) :
  signature S num_ops :=
  fun op (pfop:op < num_ops) =>
    fst (theta op pfop).




Inductive Interpretation_aux {S X U num_ops} {sig:signature S num_ops}
          {A:PAlgebra_ens U sig} {v:X->S}
          (a:assment_ens_sig (A_sigma_ens U sig A) v) :
  forall s:S, term X num_ops ->
              forall u:U,
                Inn (A_sigma_ens U sig A s) u -> Prop :=
| interpretationX :
    forall x:X,
      Interpretation_aux a (v x) (inl x::nil)
                         (proj1_sig (a x))
                         (proj2_sig (a x))
| interpretation_app :
    forall op (pfop: op < num_ops),
      let w := projT2 (fst (sig op pfop)) in
      let n := projT1 (fst (sig op pfop)) in
      let s := snd (sig op pfop) in
      forall (t:(term X num_ops)^n)
             (pft:Inn (Term_algebra_at_op sig v op pfop) t)
             (ufun:seg_fun n U),
        (forall j (pfj:j < n),
         exists pfu:Inn (A_sigma_ens U sig A
                                     (nth_prod j pfj w))
                        (ufun j pfj),
           Interpretation_aux a
                              (nth_prod j pfj w)
                              (nth_prod j pfj t)
                              (ufun j pfj)
                              pfu) ->
        let aw := map_nprod ufun in
        forall (pfaw:Inn (doms_ens U sig A op pfop) aw),
          let u := ops_sigma_ens U sig A op pfop aw pfaw in
          Interpretation_aux
            a s 
            (op_term_to_app_term
               (exist (fun c =>
                         Inn (Term_algebra_at_op sig v op pfop) c)
                      t pft))
            (proj1_sig u)
            (proj2_sig u).


Inductive Interpretation
          {S X U num_ops} {sig:signature S num_ops}
          {A:PAlgebra_ens U sig} {v:X->S}
          (a:assment_ens_sig (A_sigma_ens U sig A) v)
          (s:S) (t:term X num_ops) (u:U)
          (pfu:Inn (A_sigma_ens U sig A s) u) : Prop :=
| interpretation_aux_intro :
    Interpretation_aux a s t u pfu ->
    (forall u' (pfu':Inn (A_sigma_ens U sig A s) u'),
       Interpretation_aux a s t u' pfu' ->
       u = u') ->
    Interpretation a s t u pfu.


Lemma Interpretation_well_defined :
  forall
    {S X U num_ops} {sig:signature S num_ops}
    (v:X->S)
    (A:PAlgebra_ens U sig)
    (a:assment_ens_sig (A_sigma_ens U sig A) v) 
    (s:S) (t u:term X num_ops) (y z: U)
    (pfy:Inn (A_sigma_ens U sig A s) y)
    (pfz:Inn (A_sigma_ens U sig A s) z),
    t = u ->
    Interpretation a s t y pfy ->
    Interpretation a s u z pfz ->
    y = z.
  intros S X U num_ops sig v A a s t u y z hy hz h1 h2 h3; subst.
  destruct h3, h2.
  specialize (H0 _ hy H1); auto.
Qed.

Inductive Interpretation_dom
          {S X U num_ops} {sig:signature S num_ops}
          {A:PAlgebra_ens U sig} {v:X->S}
          (a:assment_ens_sig (A_sigma_ens U sig A) v) : 
  forall (s:S), Ensemble (term X num_ops)  :=
| interpretation_domX :
  forall x, Inn (Interpretation_dom a (v x)) (inl x::nil)
| interpretation_dom_app :
    forall op (pfop: op < num_ops),
      let w := projT2 (fst (sig op pfop)) in
      let n := projT1 (fst (sig op pfop)) in
      let s := snd (sig op pfop) in
      forall (t:(term X num_ops)^n)
             (pft:Inn (Term_algebra_at_op sig v op pfop) t)
             (ufun:seg_fun n U),
        let tm := op_term_to_app_term
                    (exist (fun c =>
                              Inn (Term_algebra_at_op sig v op pfop) c)
                           t pft) in
        (forall j (pfj:j < n),
           Inn (Interpretation_dom a (nth_prod j pfj w))
               (nth_prod j pfj t)) ->
        forall (pfin:Inn (doms_ens U sig A op pfop) (map_nprod ufun)),
          let u := ops_sigma_ens U sig A op pfop _ pfin in
          Interpretation a s tm (proj1_sig u) (proj2_sig u) ->
          Inn (Interpretation_dom a s) tm.


Lemma interpretation_ex :
  forall {S X U num_ops} {sig:signature S num_ops}
         {A:PAlgebra_ens U sig} {v:X->S}
         (a:assment_ens_sig (A_sigma_ens U sig A) v)
         (s:S) (t:term X num_ops),
    Inn (Interpretation_dom a s) t ->
    exists! u:U,
      exists (pfu:Inn (A_sigma_ens U sig A s) u),  
        Interpretation a s t u pfu.
  intros S X U num_ops sig A v a s t h1.   
  inversion h1; subst.
  exists (proj1_sig (a x)).
  red; split.
  exists (proj2_sig _).
  constructor.
  constructor.
  intros u' hu' h2.
  inversion h2; subst; auto.
  intros u h2. 
  destruct h2 as [h2 h3]. 
  inversion h3.
  inversion H; subst; auto.
  subst.
  exists (proj1_sig (ops_sigma_ens U sig A op pfop _ pfin)).
  red; split.
  exists (proj2_sig _).
  assumption.
  intros u' hu'.
  destruct hu' as [h3 h4]. 
  inversion h4.
  inversion H0.
  specialize (H2 _ (proj2_sig u) H3); subst; auto.
Qed.


Definition interpretation {S X U num_ops} {sig:signature S num_ops}
         {A:PAlgebra_ens U sig} {v:X->S}
         (a:assment_ens_sig (A_sigma_ens U sig A) v)
         (s:S) : fun_in_ens (Interpretation_dom a s) U :=
  fun (t:term X num_ops)
      (pfin:Inn (Interpretation_dom a s) t) =>
  proj1_sig (constructive_definite_description
               _
               (interpretation_ex a s t pfin)).

Lemma interpretation_compat :
  forall {S X U num_ops} {sig:signature S num_ops}
         {A:PAlgebra_ens U sig} {v:X->S}
         (a:assment_ens_sig (A_sigma_ens U sig A) v)
         (s:S) (t:term X num_ops)
         (pfin:Inn (Interpretation_dom a s) t),
    exists (pffi:Inn (A_sigma_ens U sig A s) (interpretation a s t pfin)),  
      Interpretation a s t (interpretation a s t pfin) pffi.
Proof.
  unfold interpretation; intros; apply proj2_sig.
Qed.

Lemma in_interpretation :
  forall {S X U num_ops} {sig:signature S num_ops}
         {A:PAlgebra_ens U sig} {v:X->S}
         (a:assment_ens_sig (A_sigma_ens U sig A) v)
         (s:S) (t:term X num_ops)
         (pfin:Inn (Interpretation_dom a s) t),
      Inn (A_sigma_ens U sig A s) (interpretation a s t pfin).
  intros; pose proof (interpretation_compat a s t pfin) as h1;
  simpl in h1; destruct h1; auto.
Qed.

