(* Copyright (C) 2019, Daniel Wyckoff.*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(*Version 0.3 bool2 -- certified -- Coq 8.4 pl4.*)

Require Import NaryFunctions.
Require Import FunctionAlgebrasBasics.
Require Import PartialAlgebrasBasics.
Require Import ListUtilities.
Require Import FunctionProperties.
Require Import ArithUtilities.
Require Import SetUtilities.
Require Import TypeUtilities.
Require Import TypeUtilities2.
Require Import SetUtilities1_5.
Require Import FiniteMaps.


Section PartialFunctionAlgebras.
  
  Definition function_algebra_sig : list nat :=
    1 :: 1 :: 1 :: 1 :: 2 :: nil.
    
  (*This is an interpretation of the Mal'Tsev symbols 
    (integers 0-4) into operations.
    The specific numbering of the indexing of Mal'Tsev operations 
    is arbitrary, yet conssitent with the text and previous code.*)
  Definition Jfun T : seg_fun 5 (arb_function (arb_function T)) :=
    fun i =>
      match i with
        | 0 => fun _ => unary_to_arb_function (@mzeta_to_arb T)
        | S i0 =>
          match i0 with
            | 0 => fun _ =>  unary_to_arb_function (@mtau_to_arb T)
            | S i1 =>
              match i1 with
                | 0 => fun _ => unary_to_arb_function (@mdelta_to_arb T)
                | S i2 =>
                  match i2 with
                    | 0 => fun _ => unary_to_arb_function (@mnabla_to_arb T)
                    | S i3 =>
                      match i3 with
                        | 0 =>  fun _ => binary_to_arb_function (@mstar_to_arb T)
                        | S i3 =>
                          fun pflt: S (S (S (S (S i3)))) < 5  =>
                            let pflt' :=
                                lt_S_n
                                  _ _ 
                                  (lt_S_n
                                     _ _
                                     (lt_S_n
                                        _ _
                                        (lt_S_n
                                           _ _
                                           (lt_S_n _ _ pflt)))) in
                            False_rect _ (lt_n_O _ pflt')
                      end
                  end
              end
          end
      end.

  Lemma Jfun_ar_compat :
    forall T i (pfi:i < length function_algebra_sig),
      ar (Jfun T i pfi) = nth_lt _ i pfi.
    intros T i; do 5 (destruct i; simpl; auto).
    intro; omega.
  Qed.

End PartialFunctionAlgebras.

    

(*The purpose of this section is to explore the finitary 
aspects of function generation, so that rather than look at 
sublcasses (which are infinite and usually have a finite number 
of generators), we can use the Mal'Tsev Operations to create 
relations between elements, much like the Schreier vector in 
computational group theory (and its attendant graph).*)
Section ExperimentalPointwiseGeneration.

  (*Example: [pointwise_generates {f g h i} (op h, op' i)*)
  
  Inductive pointwise_generates {T}  
            (E:Ensemble (arb_function T)) :
    Ensemble (arb_function T) -> Prop :=
| pointwise_generates_empty :
      pointwise_generates E (Empty_set _)
| mal_tsev_op1_generates :
    forall (F:Ensemble (arb_function T)),
      pointwise_generates E F ->
      forall m n
             (op:function T m -> function T n),
        mal_tsev_op1 op ->
        forall (f:function T m),
          Inn E (existT _ _ f) ->
          pointwise_generates E (Add F (existT _ _ (op f)))
| mal_tsev_op2_generates :
    forall (F:Ensemble (arb_function T)),
      pointwise_generates E F ->
      forall m n
             (op:function T n -> function T m ->
                 function T (pred (m + n))),
        mal_tsev_op2 op ->
        forall (f:function T n) (g:function T m),
          Inn E (existT _ _ f) ->
          Inn E (existT _ _ g) ->
          pointwise_generates E (Add F
                                     (existT _ _ (op f g))).

  (*Example: [pointwise_strictly_generates {h i} (op h, op' i)*)
  Inductive pointwise_strictly_generates {T} :
    Ensemble (arb_function T) ->
    Ensemble (arb_function T) -> Prop :=
  | pointwise_strictly_generates_empty :
      pointwise_strictly_generates (Empty_set _) (Empty_set _)
  | mal_tsev_op1_strictly_generates :
      forall n m
             (E F:Ensemble (arb_function T))
             (f:function T n)
             (op:function T n->function T m),
        mal_tsev_op1 op ->
        pointwise_strictly_generates E F ->
        pointwise_strictly_generates (Add E (existT _ _ f))
                                     (Add E (existT _ _ (op f)))
  | mal_tsev_op2_strictly_generates :
      forall n m
             (E F:Ensemble (arb_function T))
             (f:function T n)
             (g:function T m)
             (op:function T n->function T m->function T (pred (m+n))),
        mal_tsev_op2 op ->
        pointwise_strictly_generates E F ->
        pointwise_strictly_generates (Add
                                        (Add E (existT _ _ f))
                                        (existT _ _ g))
                                     (Add E (existT _ _ (op f g))).

    

    
  
  Lemma pointwise_generates_included1 :
    forall {T} (E E' F:Ensemble (arb_function T)),
      Included E E' ->
      pointwise_generates E F ->
      pointwise_generates E' F.
    intros T E E' F h1 h2.
    revert h1. revert E'.
    induction h2.
    intros; constructor.
    intros E' h1.
    specialize (IHh2 _ h1).
    econstructor.
    assumption.
    apply H.
    apply h1; auto.
    intros E' h1.
    specialize (IHh2 _ h1). 
    eapply mal_tsev_op2_generates; auto.
  Qed.
  

  Lemma pointwise_generates_empty1 :
    forall {T:Type} (E:Ensemble (arb_function T)),
      pointwise_generates (Empty_set _) E ->
      E = Empty_set _.
    intros T E h1.
    inversion h1; subst; auto; try contradiction.
  Qed.

  
  Lemma pointwise_generates_finite :
    forall {T:Type} (E F:Ensemble (arb_function T)),
      pointwise_generates E F ->
      Finite F.
    intros ? ? ? h1; induction h1;
    try apply Empty_is_finite; 
    try apply Add_preserves_Finite; auto.
  Qed.
  
  Lemma pointwise_generates_add1 :
    forall {T n} (E F:Ensemble (arb_function T))
           (f:function T n),
      Inn E (existT _ _ f) ->
      forall (op:function T n -> function T n),
        mal_tsev_op1 op ->
        pointwise_generates E F ->
        pointwise_generates E (Add F (existT _ _ (op f))).
    intros; constructor; auto.
  Qed.

  Lemma pointwise_generates_add2 :
    forall {T n m} (E F:Ensemble (arb_function T))
           (f:function T n) (g:function T m),
      Inn E (existT _ _ f) ->
      Inn E (existT _ _ g) ->
      forall (op:function T n -> function T m -> function T (pred (m+n))),
        mal_tsev_op2 op ->
        pointwise_generates E F ->
        pointwise_generates E (Add F (existT _ _ (op f g))).
    intros; apply mal_tsev_op2_generates; auto.
  Qed.

  
  Lemma pointwise_generates_in_impl :
    forall {T},
      FiniteT T ->
      forall (E F:Ensemble (arb_function T)) f,
        F <> Empty_set _ ->
        pointwise_generates E F ->
        Inn F f ->
        exists e,
          Inn E e /\
        ((exists (op:function T (ar e) -> function T (ar f)),
            mal_tsev_op1 op /\ 
            projT2 f = op (projT2 e)) \/
         (exists g (op:function T (ar e) -> function T (ar g) ->
                       function T (pred (ar g + ar e)))
                 (pfe:ar f = pred (ar g + ar e)),
            mal_tsev_op2 op /\
            function_transfer pfe (projT2 f) =
            op (projT2 e) (projT2 g))). 
    intros T hfin E F f he h1. revert f he.
    induction h1.
    intros; contradiction. 
    intros f0 h0 h2.
    pose proof (pointwise_generates_finite _ _ h1) as h3.    
    destruct (eq_empty_dec (arb_function_eq_dec hfin) _ h3) as [h4|h4]; subst.
    rewrite add_empty_sing in h2.
    destruct h2; subst.
    exists (existT _ _ f).
    split; auto.
    left.
    simpl.
    exists op; auto.  
    inversion h2 as [c h5|c h5]; subst.
    specialize (IHh1 f0 h4 h5); auto.
    destruct h5; subst.
    exists (existT _ _ f).
    split; auto.
    left.
    simpl.  
    exists op; auto.
    intros f0 h2 h3. 
    pose proof (pointwise_generates_finite _ _ h1) as h4.
    destruct (eq_empty_dec (arb_function_eq_dec hfin) _ h4) as [h5|h5]; subst. 
    rewrite add_empty_sing in h3.
    destruct h3; subst.
    exists (existT _ _ f).
    split; auto.
    right.
    simpl.
    exists (existT _ _ g), op, (eq_refl _); split; auto.
    inversion h3 as [c h6|c h6]; subst.
    specialize (IHh1 f0 h5 h6); auto. 
    destruct h6; subst.
    exists (existT _ _ f).
    split; auto.
    right; simpl.
    exists (existT _ _ g), op, (eq_refl _); split; auto.
  Qed.


End ExperimentalPointwiseGeneration.

(*This is merely an experiment in finite homomoprhisms on function 
algebras.*)
Section FiniteHomomorphisms.
  Inductive Fin_hom {T} :
      Ensemble (arb_function T) ->
      Ensemble (arb_function T) ->
      arb_function T -> Type :=
    fin_hom_intro1 : forall (E F:Ensemble (arb_function T))
                            (def:arb_function T),
                       Finite E -> Finite F ->
                       forall E' F',
                         pointwise_generates E' E ->
                         pointwise_generates F' F ->
                         forall (h:Fin_map E F def),
                           (forall m n (e:function T n)
                                   (pfe:n = ar (h |-> existT _ _ e)),
                              Inn E' (existT _ _ e) ->
                              forall
                                (op:function T n ->
                                    function T m)
                                (pfe':m = ar (h |-> existT _ _ (op e))),
                              Inn E (existT _ _ (op e)) ->
                              mal_tsev_op1 op ->
                              Inn F' (h |-> (existT _ _ e)) /\
                              Inn F (h |-> (existT _ _ (op e))) /\
                              h |-> (existT _ _ (op e)) =
                              existT _ _ (op (function_transfer
                                                (eq_sym pfe) 
                                                (projT2 (h |-> existT _ _ e))))) ->
                           Fin_hom E F def.
End FiniteHomomorphisms.


